# RCAIDE/Library/Methods/Propulsor/Ducted_Fan_Propulsor/translate_conditions_to_dfdc_cases.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
import  RCAIDE
from RCAIDE.Framework.Core import Data ,Units

# ---------------------------------------------------------------------------------------------------------------------- 
# Translate Conditions to DFDC Cases 
# ----------------------------------------------------------------------------------------------------------------------    
def translate_conditions_to_dfdc_cases(dfdc_analysis):
    """
    Translates flight conditions to DFDC case definitions for analysis.
    
    Parameters
    ----------
    dfdc_analysis : DFDCAnalysis
        Analysis object containing the following attributes:
            - settings.filenames.results_template : str
                Template string for naming result files
            - training : Data
                Training data parameters
                    - mach : array
                        Array of freestream Mach numbers
                    - altitude : array
                        Array of altitudes [m]
                    - tip_mach : array
                        Array of tip Mach numbers
            - geometry : DuctedFan
                Ducted fan geometry with the following attributes:
                    - cruise : Data
                        Design cruise conditions
                            - design_altitude : float
                                Design altitude [m]
                            - design_angular_velocity : float
                                Design angular velocity [rad/s]
                            - design_freestream_velocity : float
                                Design freestream velocity [m/s]
                            - design_freestream_mach : float
                                Design freestream Mach number
                    - tip_radius : float
                        Tip radius of the ducted fan [m]
    
    Returns
    -------
    None
    
    Notes
    -----
    This function creates DFDC case definitions for:
        1. The design case (using design parameters from the geometry)
        2. All combinations of Mach number, tip Mach number, and altitude specified
        in the training data
    
    For each case, it calculates:
        - Velocity based on Mach number and atmospheric conditions
        - RPM based on tip Mach number and tip radius
        - Altitude (converted to kilometers for DFDC)
    
    The US Standard Atmosphere 1976 model is used to compute atmospheric
    properties at each altitude.
    
    See Also
    --------
    RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976
    """
    # set up aerodynamic Conditions object
    template   = dfdc_analysis.settings.filenames.results_template 
    mach       = dfdc_analysis.training.mach     
    altitude   = dfdc_analysis.training.altitude        
    tip_mach   = dfdc_analysis.training.tip_mach
    ducted_fan = dfdc_analysis.geometry
    
    # first case is the design case 
    case            = Data() 
    atmosphere      = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data       = atmosphere.compute_values(ducted_fan.cruise.design_altitude)  
    rpm             = ducted_fan.cruise.design_angular_velocity/Units.rpm 
    case.tag        = template.format(ducted_fan.cruise.design_freestream_velocity,rpm,ducted_fan.cruise.design_altitude)   
    case.velocity   = ducted_fan.cruise.design_freestream_mach * atmo_data.speed_of_sound[0,0]
    case.RPM        = rpm
    case.altitude   = ducted_fan.cruise.design_altitude / 1000 # DFDC takes altitude in kilometers 
    dfdc_analysis.append_case(case)
    
    for i in range(len(mach)): 
        for j in range(len(tip_mach)):   
            for k in range(len(altitude)):     
                case            = Data() 
                atmosphere      = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
                atmo_data       = atmosphere.compute_values(altitude[k])  
                velocity        = mach[i] * atmo_data.speed_of_sound[0,0]
                tip_speed       = tip_mach[j]*atmo_data.speed_of_sound[0,0]
                omega           = tip_speed /ducted_fan.tip_radius
                rpm             = omega / Units.rpm
                case.tag        = template.format(velocity,rpm,altitude[k])  
                case.velocity   = velocity
                case.RPM        = rpm
                case.altitude   = altitude[k] / 1000
                dfdc_analysis.append_case(case) 
    return 


# RCAIDE/Framework/Analyses/Weights/__init__.py
# 

"""RCAIDE Package Setup
"""

from .Weights           import Weights
from .Conventional      import Conventional
from .Electric          import Electric
from .Hydrogen          import Hydrogen
from .Hybrid            import Hybrid



# RCAIDE/Library/Plots/Geometry/plot_Layout_of_Passenger_Accommodations.py
#  
# Created:  Mar 2025, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
from RCAIDE.Library.Methods.Geometry.LOPA.compute_layout_of_passenger_accommodations import  compute_layout_of_passenger_accommodations

# python imports 
import plotly.graph_objects as go
import numpy as  np
import os
import sys
# ----------------------------------------------------------------------------------------------------------------------
#  plot_Layout_of_Passenger_Accommodations
# ---------------------------------------------------------------------------------------------------------------------- 
def plot_layout_of_passenger_accommodations(fuselage, 
                                            save_figure    = False,
                                            save_filename  = "Vehicle_Geometry", 
                                            show_figure    = True):
    '''
    Plot aircraft layout of passenger accommodations
    '''  
    if  type(fuselage.layout_of_passenger_accommodations) != np.ndarray: 
        compute_layout_of_passenger_accommodations(fuselage)
        
    LOPA = fuselage.layout_of_passenger_accommodations
    
    fig = go.Figure() 
    # Set axes properties
    fig.update_xaxes(range=[min(LOPA[:,2]) - 1 , max(LOPA[:,2]) + 1], showgrid=False)
    fig.update_yaxes(range=[ min(LOPA[:,3]) - 1, max(LOPA[:,3])  + 1], showgrid=False)  
    fig.update_yaxes(
        scaleanchor = "x",
        scaleratio = 1,)    
        
    # Step 1: plot cabin bounds  
    # get points at x min 
    x_min_locs   =  np.where( LOPA[:,2] == min(LOPA[:,2]))[0]
    x_min        =  LOPA[x_min_locs[0],2] -  LOPA[x_min_locs[0],5]/2
    x_min_y_max  =  max(LOPA[x_min_locs,3] + LOPA[x_min_locs,6]/2 )
    x_min_y_min  =  min(LOPA[x_min_locs,3] - LOPA[x_min_locs,6]/2 ) 
    x_border_pts = [x_min, x_min] 
    y_border_pts = [x_min_y_min, x_min_y_max] 

    # get points at y max 
    y_max_locs   =  np.where( LOPA[:,3] == max(LOPA[:,3]))[0]
    y_max        =  LOPA[y_max_locs[0],3] + LOPA[y_max_locs[0],6]/2 
    y_max_x_max  =  max(LOPA[y_max_locs,2] + LOPA[y_max_locs[0],5]/2)
    y_max_x_min  =  min(LOPA[y_max_locs,2] - LOPA[y_max_locs[0],5]/2) 
    x_border_pts.append(y_max_x_min)
    x_border_pts.append(y_max_x_max)
    y_border_pts.append(y_max)
    y_border_pts.append(y_max) 

    # get points at x max 
    x_max_locs   =  np.where( LOPA[:,2] == max(LOPA[:,2]))[0]
    x_max        =  LOPA[x_max_locs[0],2] + LOPA[x_max_locs[0],5]/2
    x_max_y_max  =  max(LOPA[x_max_locs,3] + LOPA[x_max_locs,6]/2)
    x_max_y_min  =  min(LOPA[x_max_locs,3] - LOPA[x_max_locs,6]/2)  
    x_border_pts.append(x_max)
    x_border_pts.append(x_max)
    y_border_pts.append(x_max_y_max)
    y_border_pts.append(x_max_y_min)
    
    # get points at y min  
    y_min_locs   =  np.where( LOPA[:,3] == min(LOPA[:,3]))[0]
    y_min        =  LOPA[y_min_locs[0],3] - LOPA[y_min_locs[0],6]/2 
    y_min_x_max  =  max(LOPA[y_min_locs,2] + LOPA[y_min_locs[0],5]/2)
    y_min_x_min  =  min(LOPA[y_min_locs,2] - LOPA[y_min_locs[0],5]/2)
    x_border_pts.append(y_min_x_max)  
    x_border_pts.append(y_min_x_min)
    y_border_pts.append(y_min)
    y_border_pts.append(y_min)    
    
    # loop through points and determine if there are duplicates
    y_border_pts = np.array(y_border_pts)
    x_border_pts = np.array(x_border_pts)
    # cut where y is negative
    port_idxs  =  np.where(y_border_pts<0)[0]
    starboard_x_points = np.delete(x_border_pts, port_idxs) 
    starboard_y_points = np.delete(y_border_pts, port_idxs)
     
    fig.add_trace(go.Scatter(
        x=starboard_x_points, 
        y=starboard_y_points, 
        mode='lines',
        name='Line 1',
        line_color="darkblue", 
        fill=None 
    ))
    
    fig.add_trace(go.Scatter(
        x=starboard_x_points, 
        y=-starboard_y_points, 
        mode='lines',
        name='Line 2',
        line_color="darkblue", 
        fill='tonexty'
    ))
    
     
    # Step 2: plot seats
    economoy_seat_colors = ["steelblue", "deepskyblue", "skyblue" ]
    business_seat_colors = ["seagreen", "mediumseagreen", "lightseagreen" ]
    first_seat_colors    = ["indianred", "lightcoral", "lightpink" ]
    lavatory_color       = ["sandybrown"]
   
    for i in range(len(LOPA)):
        x_center     = LOPA[i,2]
        y_center     = LOPA[i,3] 
        s_l          = LOPA[i,5]
        s_w          = LOPA[i,6] 
        F_c          = LOPA[i,7]       
        B_c          = LOPA[i,8] 
        E_c          = LOPA[i,9]
        seat         = LOPA[i,10] 
        Em_row       = LOPA[i,11]
        Gal_Lav      = LOPA[i,12]
        
        if E_c == 1.0: 
            if seat == 1.0: 
                x0_pt = x_center - s_l / 2
                x1_pt = x_center + s_l / 2
                y0_pt = y_center - s_w / 2
                y1_pt = y_center + s_w / 2
                
                if Em_row == 1.0:
                    
                    fig.add_shape(type="rect",
                        x0=x0_pt, y0=y0_pt, x1=x1_pt, y1=y1_pt,
                        line=dict(
                            color=economoy_seat_colors[0],
                            width=2,
                        ),
                        fillcolor= economoy_seat_colors[1],
                    )
                    
                else: 
                    fig.add_shape(type="rect",
                        x0=x0_pt, y0=y0_pt, x1=x1_pt, y1=y1_pt,
                        line=dict(
                            color=economoy_seat_colors[0],
                            width=2,
                        ),
                        fillcolor=economoy_seat_colors[2],
                    )
        if B_c == 1: 
            if seat == 1: 
                x0_pt = x_center - s_l / 2
                x1_pt = x_center + s_l / 2
                y0_pt = y_center - s_w / 2
                y1_pt = y_center + s_w / 2
                
                if Em_row == 1:
                    
                    fig.add_shape(type="rect",
                        x0=x0_pt, y0=y0_pt, x1=x1_pt, y1=y1_pt,
                        line=dict(
                            color=business_seat_colors[0],
                            width=2,
                        ),
                        fillcolor=business_seat_colors[1],
                    )
                    
                else:
                    
                    fig.add_shape(type="rect",
                        x0=x0_pt, y0=y0_pt, x1=x1_pt, y1=y1_pt,
                        line=dict(
                            color=business_seat_colors[0],
                            width=2,
                        ),
                        fillcolor=business_seat_colors[2],
                    )                    
    
        if F_c == 1: 
            if seat == 1: 
                x0_pt = x_center - s_l / 2
                x1_pt = x_center + s_l / 2
                y0_pt = y_center - s_w / 2
                y1_pt = y_center + s_w / 2
                
                if Em_row == 1:
                    
                    fig.add_shape(type="rect",
                        x0=x0_pt, y0=y0_pt, x1=x1_pt, y1=y1_pt,
                        line=dict(
                            color=first_seat_colors[0],
                            width=2,
                        ),
                        fillcolor=first_seat_colors[1],
                    )
                    
                else:
                    
                    fig.add_shape(type="rect",
                        x0=x0_pt, y0=y0_pt, x1=x1_pt, y1=y1_pt,
                        line=dict(
                            color=first_seat_colors[0],
                            width=2,
                        ),
                        fillcolor=first_seat_colors[2],
                    )    
     
        if Gal_Lav == 1:
            x0_pt = x_center - s_l / 2
            x1_pt = x_center + s_l / 2
            y0_pt = y_center - s_w / 2
            y1_pt = y_center + s_w / 2
             
            fig.add_shape(type="rect",
                x0=x0_pt, y0=y0_pt, x1=x1_pt, y1=y1_pt,
                line=dict(
                    color=lavatory_color[0],
                    width=2,
                ),
                fillcolor=lavatory_color[0],
            )

    # Use the first path from sys.path
    save_filename = os.path.join(sys.path[0], save_filename)
    if save_figure:
        fig.write_image(save_filename + ".png")
        
    if show_figure:
        fig.write_html( save_filename + '.html', auto_open=True) 
     
    return 


# RCAIDE/Library/Plots/Aerodynamics/plot_lift_distribution.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt 
import numpy as np 
# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_lift_distribution(results,
                           save_figure = False,
                           save_filename = "Lift_Distribution",
                           file_type = ".png",
                           width = 11, height = 7):
    """
    Generate plots of spanwise lift distribution for lifting surfaces.

    Parameters
    ----------
    results : Data
        Mission results data structure containing:
            - results.segments.conditions.aerodynamics.coefficients.lift with fields:
                - inviscid_wings_sectional : array
                    Sectional lift coefficients at control points

    save_figure : bool, optional
        Save figure to file if True, default False

    save_filename : str, optional
        Base name for saved figure files, default "Lift_Distribution"

    file_type : str, optional
        File extension for saved figure, default ".png"

    width : float, optional
        Figure width in inches, default 11

    height : float, optional
        Figure height in inches, default 7

    Returns
    -------
    fig : matplotlib.figure.Figure
        Figure showing spanwise lift distribution

    Notes
    -----
    Creates figures showing:
        - Sectional lift coefficient (CLy) vs spanwise location
        - Separate plot for each timestep in each segment
        - Different wings distinguished by line colors:
            - Blue: Main wings
            - Red: Horizontal tails
            - Black: Other surfaces

    **Definitions**

    'Sectional Lift Coefficient'
        Non-dimensional lift force per unit span
    
    'Control Points'
        Points where circulation/lift is evaluated

    See Also
    --------
    RCAIDE.Library.Plots.Common.set_axes : Standardized axis formatting
    RCAIDE.Library.Plots.Common.plot_style : RCAIDE plot styling
    """   

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                      'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
    
    VD         = results.segments[0].analyses.aerodynamics.vehicle.vortex_distribution	 	
    n_w        = VD.n_w
    b_sw       = np.concatenate(([0],np.cumsum(VD.n_sw)))

    img_idx    = 1
    seg_idx    = 1
    for segment in results.segments.values():   	
        num_ctrl_pts = len(segment.conditions.frames.inertial.time)	
        for ti in range(num_ctrl_pts):  
            cl_y = segment.conditions.aerodynamics.coefficients.lift.induced.spanwise[ti] 
            line = ['-b','-b','-r','-r','-k']
            fig  = plt.figure(save_filename + '_' + str(img_idx))
            fig.set_size_inches(8,8)  
            fig.set_size_inches(width,height)     
            axes = plt.subplot(1,1,1)
            for i in range(n_w): 
                y_pts = VD.Y_SW[b_sw[i]:b_sw[i+1]]
                z_pts = cl_y[b_sw[i]:b_sw[i+1]]
                axes.plot(y_pts, z_pts, line[i] ) 
            axes.set_xlabel("Spanwise Location (m)")
            axes.set_title('$C_{Ly}$')  

            if save_figure: 
                plt.savefig( save_filename + '_' + str(img_idx) + file_type) 	
            img_idx += 1
        seg_idx +=1

    return fig 


# RCAIDE/Library/Attributes/Atmospheres/Earth/US_Standard_1976.py
# 
#
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 
 
from RCAIDE.Framework.Core import Data , Units
from RCAIDE.Library.Attributes.Gases import Air
from RCAIDE.Library.Attributes.Atmospheres import Atmosphere
from RCAIDE.Library.Attributes.Planets import Earth 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  US_Standard_1976 Class
# ----------------------------------------------------------------------------------------------------------------------   
class US_Standard_1976(Atmosphere):
    """
    Implementation of the 1976 U.S. Standard Atmosphere model.

    This class provides a complete implementation of the U.S. Standard Atmosphere (1976),
    defining atmospheric properties at various altitudes based on empirical data.

    Attributes
    ----------
    fluid_properties : Air
        Air properties object containing gas characteristics
    planet : Earth
        Earth properties object containing planetary parameters
    breaks : Data
        Container for atmospheric property breakpoints
            - altitude : array
                Geopotential altitude points in kilometers, ranging from -2 km to 84.852 km
            - temperature : array
                Temperature values at break points in Kelvin
            - pressure : array
                Pressure values at break points in Pascal
            - density : array
                Density values at break points in kg/m³

    Notes
    -----
    The U.S. Standard Atmosphere 1976 is a model defining atmospheric properties 
    up to 1000 km. This implementation covers the range from -2 km to 84.852 km, 
    which encompasses the primary region of interest for most aircraft operations.

    The model divides the atmosphere into layers with different temperature gradients
    and provides standardized values for Temperature, Pressure, Density, Speed of sound, and Viscosity.

    **Major Assumptions**
        * Hydrostatic equilibrium
        * Perfect gas behavior
        * Homogeneous composition below 80 km
        * Gravity variation with geometric height

    References
    ----------
    [1] NOAA, NASA, USAF, U.S. Standard Atmosphere, 1976 (1976). Retrieved December 30, 2024, from https://www.ngdc.noaa.gov/stp/space-weather/online-publications/miscellaneous/us-standard-atmosphere-1976/us-standard-atmosphere_st76-1562_noaa.pdf. 

    See Also
    --------
    RCAIDE.Library.Attributes.Atmospheres.Earth.Constant_Temperature : Simplified constant temperature model
    """
    
    def __defaults__(self):
        """This sets the default values at breaks in the atmosphere. 
    
        Assumptions:
            None
        
        Source:
            U.S. Standard Atmosphere (1976 version) 
        """
        self.tag = ' U.S. Standard Atmosphere (1976)'

        # break point data: 
        self.fluid_properties   = Air()
        self.planet             = Earth()
        self.breaks             = Data()
        self.breaks.altitude    = np.array( [-2.00    , 0.00,     11.00,      20.00,      32.00,      47.00,      51.00,      71.00,      84.852]) * Units.km     # m, geopotential altitude
        self.breaks.temperature = np.array( [301.15   , 288.15,   216.65,     216.65,     228.65,     270.65,     270.65,     214.65,     186.95])      # K
        self.breaks.pressure    = np.array( [127774.0 , 101325.0, 22632.1,    5474.89,    868.019,    110.906,    66.9389,    3.95642,    0.3734])      # Pa
        self.breaks.density     = np.array( [1.47808e0, 1.2250e0, 3.63918e-1, 8.80349e-2, 1.32250e-2, 1.42753e-3, 8.61606e-4, 6.42099e-5, 6.95792e-6])  # kg/m^3


# RCAIDE/Library/Attributes/Gases/__init__.py
# 

"""
This module provides gas handling capabilities for RCAIDE. 

It includes classes for managing different types of gases and their thermodynamic properties, 
with specific implementations for air, carbon dioxide (CO2), and steam. The Gas class serves 
as the base class for all gas implementations.

See Also
--------
RCAIDE.Library.Attributes.Cryogens : Related module for cryogenic fluid handling
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Gas    import Gas
from .Air    import Air
from .CO2    import CO2
from .Steam  import Steam


# RCAIDE/Library/Attributes/Propellants/JP7.py
# 
#
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from .Propellant import Propellant 

# ---------------------------------------------------------------------------------------------------------------------- 
#  JP7 Propellant Class
# ----------------------------------------------------------------------------------------------------------------------  
class JP7(Propellant):
    """
    A class representing JP-7 high thermal stability jet fuel properties. This specialized 
    fuel was developed for high-speed aircraft operating at elevated temperatures.

    Attributes
    ----------
    tag : str
        Identifier for the propellant ('JP7')
    reactant : str
        Oxidizer used for combustion ('O2')
    density : float
        Fuel density in kg/m³ (803.0)
    specific_energy : float
        Specific energy content in J/kg (43.50e6)
    energy_density : float
        Energy density in J/m³ (34930.5e6)
    stoichiometric_fuel_to_air : float
        Stoichiometric fuel-to-air ratio (0.0674)
    temperatures : Data
        Critical temperatures in K
            - flash : float
                Flash point (333.15)
            - autoignition : float
                Autoignition temperature (555.15)
            - freeze : float
                Freezing point (514.15)

    Notes
    -----
    JP-7 is a specialized jet fuel developed for the SR-71 Blackbird and similar 
    high-speed aircraft. It features high thermal stability and low volatility for 
    operation at elevated temperatures.

    **Definitions**
    
    'Flash Point'
        Lowest temperature at which fuel vapors will ignite (60°C, significantly 
        higher than conventional jet fuels)
    
    'Thermal Stability'
        Resistance to thermal decomposition at elevated temperatures
    
    'Stoichiometric Fuel-to-Air Ratio'
        Ideal fuel-to-air mass ratio for complete combustion

    **Major Assumptions**
        * Properties are for standard temperature and pressure conditions (15C, 1atm)

    References
    ----------
    [1] Roberts, K. (2008). ANALYSIS AND DESIGN OF A HYPERSONIC SCRAMJET ENGINE WITH A STARTING MACH NUMBER OF 4.00 (thesis). UTA. University of Texas at Austin. Retrieved December 30, 2024, from https://arc.uta.edu/publications/td_files/Kristen%20Roberts%20MS.pdf. 
    """

    def __defaults__(self):
        """This sets the default values.

        Assumptions:
            None

        Source: 
            http://arc.uta.edu/publications/td_files/Kristen%20Roberts%20MS.pdf 
        """    
        self.tag                        = 'JP7'
        self.reactant                   = 'O2'
        self.density                    = 803.0                          # kg/m^3 (15 C, 1 atm)
        self.specific_energy            = 43.50e6                        # J/kg
        self.energy_density             = 34930.5e6                      # J/m^3
        self.stoichiometric_fuel_to_air = 0.0674            

        # critical temperatures
        self.temperatures.flash        = 333.15                 # K
        self.temperatures.autoignition = 555.15                 # K
        self.temperatures.freeze       = 514.15                 # K

        self.stoichiometric_fuel_air_ratio = 0         # [-] Stoichiometric Fuel to Air ratio
        self.heat_of_vaporization          = 0         # [J/kg] Heat of vaporization at standard conditions
        self.temperature                   = 0         # [K] Temperature of fuel
        self.pressure                      = 0         # [Pa] Pressure of fuel
        self.fuel_surrogate_S1             = {} # [-] Mole fractions of fuel surrogate species
        self.kinetic_mechanism             = '' # [-] Kinetic mechanism for fuel surrogate species
        self.oxidizer                      = ''


# RCAIDE/Methods/Powertrain/Sources/Fuel_Cell_Stacks/Proton_Exchange_Membrane/compute_fuel_cell_performance.py
#  
# Created: Jan 2025, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import  numpy as  np
from scipy.optimize import minimize_scalar

# ----------------------------------------------------------------------------------------------------------------------
#  Model to Compute Fuel Cell Performance
# ----------------------------------------------------------------------------------------------------------------------
def compute_fuel_cell_performance(fuel_cell_stack, state, bus, coolant_lines, t_idx, delta_t):
    """
    Computes the performance of a PEM fuel cell stack.
    
    Parameters
    ----------
    fuel_cell_stack : RCAIDE.Components.Energy.Converters.Fuel_Cell_Stack
        The fuel cell stack object containing cell properties and configuration
    state : RCAIDE.Framework.Mission.Common.State
        Container for mission segment conditions
    bus : RCAIDE.Components.Energy.Distribution.Electric_Bus
        The electric bus to which the fuel cell stack is connected
    coolant_lines : list
        List of coolant line components for thermal management
    t_idx : int
        Current time index in the simulation
    delta_t : float
        Time step size [s]
         
    Returns
    -------
    stored_results_flag : bool
        Flag indicating that results have been stored for potential reuse
    stored_fuel_cell_stack_tag : str
        Tag identifier of the fuel cell stack with stored results
    
    Notes
    -----
    This function implements a detailed electrochemical model for PEM fuel cells that accounts
    for activation losses, ohmic losses, and concentration losses. It iteratively determines
    the current density required to match the power demand from the electric bus.
    
    The function also models the compressor-expander module (CEM) that supplies air to the
    fuel cell, accounting for parasitic power losses.
    
    Temperature effects are considered through a thermal model that updates the stack
    temperature based on heat generation.
    
    **Major Assumptions**
        * Uniform temperature distribution across all cells
        * Hydrogen is supplied at constant pressure
        * Air composition is standard atmospheric (21% oxygen)
        * Water management is ideal (no flooding or membrane dehydration)
    
    **Theory**
    
    The PEM fuel cell voltage is calculated as:
    
    .. math::
        V_{cell} = E_{cell} - \eta_{act} - \eta_{ohmic} - \eta_{conc} - degradation
    
    where:
        - E_{cell} is the Nernst potential
        - η_{act} is the activation loss
        - η_{ohmic} is the ohmic loss
        - η_{conc} is the concentration loss
        - degradation accounts for cell degradation over time
    
    The thermal model uses:
    
    .. math::
        \\frac{dT}{dt} = \\frac{Q_{gross}}{m_{stack} \\cdot c_p}
    
    where Q_{gross} is the heat generated by the stack.
    
    References
    ----------
    [1] O'Hayre, R., Cha, S. W., Colella, W., & Prinz, F. B. (2016). Fuel Cell Fundamentals (3rd ed.). John Wiley & Sons.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Proton_Exchange_Membrane.evaluate_PEM
    RCAIDE.Library.Methods.Powertrain.Converters.Fuel_Cells.Proton_Exchange_Membrane.evaluate_CEM
    """
    # ---------------------------------------------------------------------------------
    # Unpack flight conditions 
    # ---------------------------------------------------------------------------------
    M0 = state.conditions.freestream.mach_number
    P0 = state.conditions.freestream.pressure
    T0 = state.conditions.freestream.temperature
    a  = state.conditions.freestream.speed_of_sound
    R  = 287
 
    # Compute the working fluid properties 
    gamma    =  (a ** 2) /( T0 * R)

    # Compute the stagnation quantities from the input static quantities
    stagnation_pressure    = P0*((1.+(gamma-1.)/2.*M0*M0 )**(gamma/(gamma-1.))) 
    stagnation_temperature = T0*(1.+((gamma-1.)/2.*M0*M0))
     
    # fuel cell properties  
    fuel_cell         = fuel_cell_stack.fuel_cell 
    n_series          = fuel_cell_stack.electrical_configuration.series
    n_parallel        = fuel_cell_stack.electrical_configuration.parallel 
    n_total           = n_series*n_parallel  
     
    # Compute Bus electrical properties  
    bus_conditions              = state.conditions.energy[bus.tag] 
    P_bus                       = bus_conditions.power_draw
    bus_config                  = bus.fuel_cell_stack_electric_configuration 
    P_module                    = P_bus  /len(bus.fuel_cell_stacks)
    P_cell                      = P_module[t_idx]/n_total 
     
    # Compute fuel_cell_stack Conditions 
    fuel_cell_stack_conditions = state.conditions.energy[bus.tag].fuel_cell_stacks[fuel_cell_stack.tag]
    
    # append atmospheric conditions to fuel cell 
    fuel_cell_stack_conditions.fuel_cell.stagnation_temperature[t_idx] = stagnation_temperature[t_idx]
    fuel_cell_stack_conditions.fuel_cell.stagnation_pressure[t_idx]    = stagnation_pressure[t_idx]
    
    i      = 1
    diff_P =  10
    alpha  =  0.05
    while  abs(diff_P) > 1E-8: 
        fuel_cell_stack_conditions.fuel_cell.current_density[t_idx] = i
        m_dot_H2, V_fuel_cell, P_fuel_cell, gross_power, gross_heat, compressor_power, mdot_air_in, mdot_air_out, expander_power =  evaluate_PEM(fuel_cell_stack,fuel_cell_stack_conditions, t_idx) 
        diff_P =  P_cell - P_fuel_cell
        i += diff_P*alpha
        
    # ---------------------------------------------------------------------------------------------------     
    # Future State 
    # --------------------------------------------------------------------------------------------------- 
    if t_idx != state.numerics.number_of_control_points-1:   
        dT_dt              =  gross_heat / (fuel_cell.mass *fuel_cell.specific_heat_capacity) 
        fuel_cell_stack_conditions.fuel_cell.stack_temperature[t_idx+1,0]  = fuel_cell_stack_conditions.fuel_cell.stack_temperature[t_idx, 0] + dT_dt*delta_t[t_idx]  
    
    I_cell  = P_fuel_cell / V_fuel_cell
    I_stack = I_cell * n_parallel
    if bus_config == 'Series':
        bus_conditions.current_draw[t_idx] = I_stack  
    elif bus_config  == 'Parallel': 
        bus_conditions.current_draw[t_idx] = I_stack * len(bus.fuel_cell_stacks)  
     
    
    fuel_cell_stack_conditions.power[t_idx]                                      = P_fuel_cell * n_total
    fuel_cell_stack_conditions.current[t_idx]                                    = I_stack
    fuel_cell_stack_conditions.voltage_open_circuit[t_idx]                       = V_fuel_cell *  n_series # assumes no losses
    fuel_cell_stack_conditions.voltage_under_load[t_idx]                         = V_fuel_cell *  n_series
    fuel_cell_stack_conditions.fuel_cell.voltage_open_circuit[t_idx]             = V_fuel_cell   # assumes no losses
    fuel_cell_stack_conditions.fuel_cell.voltage_under_load[t_idx]               = V_fuel_cell
    fuel_cell_stack_conditions.fuel_cell.power[t_idx]                            = P_fuel_cell
    fuel_cell_stack_conditions.fuel_cell.current[t_idx]                          = P_fuel_cell / V_fuel_cell  
    fuel_cell_stack_conditions.H2_mass_flow_rate[t_idx]                          = m_dot_H2 * n_total
    fuel_cell_stack_conditions.fuel_cell.inlet_H2_mass_flow_rate[t_idx]          = m_dot_H2
    fuel_cell_stack_conditions.fuel_cell.inlet_H2_mass_flow_rate[t_idx]          = m_dot_H2
    fuel_cell_stack_conditions.fuel_cell.inlet_air_mass_flow_rate[t_idx]         = mdot_air_in
    fuel_cell_stack_conditions.fuel_cell.outlet_air_mass_flow_rate[t_idx]        = mdot_air_out
    
    stored_results_flag            = True
    stored_fuel_cell_stack_tag     = fuel_cell_stack.tag  

    return  stored_results_flag, stored_fuel_cell_stack_tag

def evaluate_PEM(fuel_cell_stack,fuel_cell_conditions,t_idx):
    """
    Determines the fuel cell state of the PEM fuel cell 

    Parameters: 
    ----------
    i: float 
        The current density to evaluate the cell voltage at (A/cm^2)
    stack_temperature: Temperature 
        Temperature of the fuel cell
    P_H2_input: Pressure
        Pressure of the hydrogen stream 
    FC_air_p: Pressure
        Pressure of the hydrogen air entering the fuel cell
    RH: float 
        Relative humidity of the air and hydrogen streams (0-1)
    air_excess_ratio: float 
        Air excess ratio (above stoichometric)
    thermo_state_in: ThermoState 
        Conditions of air entering the compressor
    degradation: float 
        The voltage drop due to degradation (V)
        optional, defaults to 0 V 

    Returns: 
    ----------
    float: 
        Hydrogen Mass flow (kg/s)
    float: 
        Net Power (W)
    float: 
        Gross power (W)
    float: 
        Gross heat (W)
    float: 
        Compressor Power (W)
    float: 
        mdot_air_in (kg/s)
    """
    fuel_cell        = fuel_cell_stack.fuel_cell 
    i                = fuel_cell_conditions.fuel_cell.current_density[t_idx]  
    air_excess_ratio = fuel_cell.air_excess_ratio 
    
    fuel_cell_conditions.fuel_cell.pressure_drop[t_idx] = calculate_P_drop_stack(fuel_cell_stack,i)
    if fuel_cell.type == "LT":
        fuel_cell_conditions.fuel_cell.humidifier.pressure_drop[t_idx] = calculate_P_drop_hum(fuel_cell_stack,i) 
    else: 
        fuel_cell_conditions.fuel_cell.humidifier.pressure_drop[t_idx] = 0 
    
    mdot_air_in     = i * fuel_cell.interface_area * fuel_cell.O2_molar_mass / (4 * fuel_cell.Faraday_constant * fuel_cell.O2_mass_frac) * air_excess_ratio
    mdot_H2         = i * fuel_cell.interface_area / (2 * fuel_cell.Faraday_constant) * fuel_cell.H2_molar_mass
    voltage, V_loss = calculate_voltage(i, fuel_cell_stack,fuel_cell_conditions,t_idx)
    gross_power     = voltage * i * fuel_cell.interface_area
    gross_heat      = V_loss * i * fuel_cell.interface_area
    
    # evalaute compressor_expander_module
    fuel_cell_conditions.fuel_cell.inlet_air_mass_flow_rate[t_idx] = mdot_air_in
    compressor_power, mdot_air_out, expander_power = evaluate_CEM(fuel_cell_stack, fuel_cell_conditions,t_idx)
    parasitic_power  = fuel_cell.gamma_para * gross_power
    net_power        = gross_power - compressor_power - parasitic_power
    
    return mdot_H2, voltage, net_power, gross_power, gross_heat, compressor_power, mdot_air_in, mdot_air_out, expander_power

def evaluate_CEM(fuel_cell_stack,fuel_cell_conditions,t_idx):
    """
    Evaluates the power required by the CEM (compressor-expander module)

    Parameters: 
    ----------
    FC_air_p: float 
        Air pressure after the humidifier entering the fuel cell (bar)
    thermo_state_in: ThermoState 
        Air thermostate entering the compressor
    mdot_air_in: float 
        Mass flow of air entering the fuel cell and compressor (kg/s)
    air_excess_ratio: float 
        Oxygen excess ratio (above stoichometric)
    p_drop_hum: float 
        The pressure drop (bar) through the humidifier 
    pressure_drop: float 
        The pressure drop (bar) through the fuel cell

    Returns: 
    ----------
    float: 
        The power required to run the CEM at the given operating conditions (W)
    """
    Cp                = 1004
    gam               = 1.4
    
    fuel_cell         = fuel_cell_stack.fuel_cell 
    CEM               = fuel_cell.compressor_expander_module 
    Tt_in             = fuel_cell_conditions.fuel_cell.stagnation_temperature[t_idx, 0] 
    Pt_in             = fuel_cell_conditions.fuel_cell.stagnation_pressure[t_idx, 0]  
    mdot_air_in       = fuel_cell_conditions.fuel_cell.inlet_air_mass_flow_rate[t_idx, 0]
    p_drop_hum        = fuel_cell_conditions.fuel_cell.humidifier.pressure_drop[t_idx, 0]
    pressure_drop     = fuel_cell_conditions.fuel_cell.pressure_drop[t_idx, 0]
    FC_air_p          = fuel_cell.rated_air_pressure  
    air_excess_ratio  = fuel_cell.air_excess_ratio 
    
    p_air_FC          = FC_air_p  + p_drop_hum
    comp_p_req        = mdot_air_in * Cp* Tt_in * ( ((p_air_FC + p_drop_hum) / Pt_in) **((gam - 1) / gam)   - 1) / CEM.compressor_efficiency
    input_p           = comp_p_req / CEM.motor_efficiency 
    p_exp             = p_air_FC - pressure_drop -p_drop_hum
    Tt_exp            = Tt_in * (p_exp / Pt_in) ** ((gam - 1) / gam)
    mdot_air_out      = mdot_air_in -  mdot_air_in /air_excess_ratio * 0.233
    exp_p_ext         = mdot_air_out * Cp * Tt_exp * (1 - (Pt_in / p_exp) ** ((gam - 1) / gam)) * CEM.expander_efficiency
    output_p          = exp_p_ext * CEM.generator_efficiency
    p_req             = input_p - output_p

    fuel_cell_conditions.fuel_cell.outlet_air_pressure[t_idx]              = FC_air_p  + p_drop_hum     
    fuel_cell_conditions.fuel_cell.compressor_inlet_pressure[t_idx]        = Pt_in 
    fuel_cell_conditions.fuel_cell.compressor_pressure_ratio[t_idx]        = (p_air_FC + p_drop_hum) / Pt_in 
    fuel_cell_conditions.fuel_cell.compressor_inlet_mass_flow_rate[t_idx]  = mdot_air_in 
    fuel_cell_conditions.fuel_cell.compressor_power[t_idx]                 = comp_p_req
    fuel_cell_conditions.fuel_cell.expander_outlet_mass_flow_rate[t_idx]   = mdot_air_out 
    fuel_cell_conditions.fuel_cell.expander_inlet_pressure[t_idx]          = p_exp 
    fuel_cell_conditions.fuel_cell.expander_pressure_ratio[t_idx]          = Pt_in / p_exp
    fuel_cell_conditions.fuel_cell.expander_power[t_idx]                   = exp_p_ext 
    fuel_cell_conditions.fuel_cell.motor_compressor_power[t_idx]           = input_p
    fuel_cell_conditions.fuel_cell.expander_generator_power[t_idx]         = output_p 
    fuel_cell_conditions.fuel_cell.compressor_expander_module_power[t_idx] = p_req
       
    return p_req, mdot_air_out, exp_p_ext 
 
def calculate_voltage(i, fuel_cell_stack,fuel_cell_conditions,t_idx):
    """
    Calculates the output voltage of the fuel cell by subtracting the activation,
    ohmic, and concentration voltage losses from the reversible Nernst voltage, E_cell.

    Parameters:
    ----------
    i: float 
        The current density to evaluate the cell voltage at (A/m^2)
    stack_temperature: float 
        Temperature of the fuel cell (K)
    P_H2_input: float 
        Pressure of the hydrogen stream (bar)
    P_air: float 
        Pressure of the incoming air (bar)
    RH: float 
        Relative humidity of the air and hydrogen streams (0-1)
    air_excess_ratio: float 
        Air excess ratio (above stoichometric)
    pressure_drop: float 
        Pressure drop of the fuel cell at the rated current density (bar)
    degradation: float 
        The percent of maximumm degradation to evaluate divided by 100
        optional, defaults to 0 (0% of fuel_cell.maximum_deg)

    Returns:
    ----------
    float: 
        Output voltage of the fuel cell (V)
    """
    
    # unpack
    fuel_cell          = fuel_cell_stack.fuel_cell 
    stack_temperature  = fuel_cell_conditions.fuel_cell.stack_temperature[t_idx, 0]      
    P_H2_input         = fuel_cell.rated_H2_pressure   
    P_air              = fuel_cell.rated_air_pressure
    RH                 = fuel_cell.oxygen_relative_humidity 
    air_excess_ratio   = fuel_cell.air_excess_ratio 
    pressure_drop      = fuel_cell_conditions.fuel_cell.pressure_drop[t_idx, 0]      
    degradation        = fuel_cell_conditions.fuel_cell.degradation[t_idx, 0]    
    
    P_O2   = calculate_P_O2(fuel_cell_stack,P_air, stack_temperature, RH, air_excess_ratio, pressure_drop, i)
    P_H2   = calculate_P_H2(fuel_cell_stack,P_H2_input, stack_temperature, RH, i)
    E_cell = calculate_E_cell(fuel_cell_stack,stack_temperature, P_H2, P_O2)
    
    if fuel_cell.type == "LT": 
        eta_ohmic  = calculate_ohmic_losses_LT(fuel_cell_stack,stack_temperature, i) 
        eta_conc   = calculate_concentration_losses_LT(fuel_cell_stack,stack_temperature, P_O2, RH, air_excess_ratio, pressure_drop, i)
    elif fuel_cell.type == "HT": 
        eta_ohmic = calculate_ohmic_losses_HT(fuel_cell_stack,stack_temperature, i) 
        eta_conc  = calculate_concentration_losses_HT(fuel_cell_stack,stack_temperature, P_O2, RH, air_excess_ratio, pressure_drop, i) 
    eta_act = calculate_activation_losses(fuel_cell_stack,stack_temperature, P_O2, i) 

    # Calculate the output voltage of the fuel cell
    V_cell = E_cell - eta_act - eta_ohmic - eta_conc - degradation * fuel_cell.maximum_deg
    V_loss = eta_act + eta_ohmic + eta_conc + degradation * fuel_cell.maximum_deg
    return V_cell, V_loss
 
def set_rated_current_density(fuel_cell_stack, rated_current_density, rated_power_density): 
    """
    Sets the rated cd and pd of the fuel cell system 

    Parameters: 
    ----------
    rated_CD: float
        The current density (A/cm2) to set 
    rated_power_density: float 
        The power density (W/cm2) to set 
    """
    fuel_cell                       = fuel_cell_stack.fuel_cell
    fuel_cell.rated_current_density = rated_current_density 
    fuel_cell.rated_power_density   = rated_power_density
    
    return
 
def calculate_P_drop_hum(fuel_cell_stack, i):
    """
    Calculates the pressure drop across the humidifier for the rated current 

    Parameters: 
    ----------
    i: float 
        Current Density (Acm2) to calculate the pressure drop 

    Returns: 
    ----------
    float 
        The Humidifier pressure drop in bar
    """
    fuel_cell   = fuel_cell_stack.fuel_cell
    P_drop = (i/fuel_cell.rated_current_density) ** 2 * fuel_cell.rated_p_drop_hum
    return P_drop
 
def calculate_P_drop_stack(fuel_cell_stack, i):
    """
    Calculates the pressure drop across the stack for the rated current 

    Parameters: 
    ----------
    i: float 
        Current Density (Acm2) to calculate the pressure drop

    Returns:
    ---------- 
    float 
        The stack pressure drop in bar 
    """
    fuel_cell   = fuel_cell_stack.fuel_cell
    P_drop = (i/fuel_cell.rated_current_density) ** 2 * fuel_cell.rated_p_drop_fc
    return P_drop

def calculate_P_O2(fuel_cell_stack, P_air, stack_temperature, RH, air_excess_ratio, P_drop, i): 
    """
    Calculate the partial pressure of oxygen at the fuel cell cathode 

    Parameters: 
    ----------
    P_air: float 
        The input air pressure to the fuel cell (bar)
    stack_temperature: float 
        The fuel cell stack temperature (K)
    RH: float 
        The relative humidity of the oxygen stream (0-1)
    air_excess_ratio: float 
        The oxygen excess ratio (above stoichometric)
    P_drop: float 
        The pressure drop (bar) across the fuel cell cathode 
    i: float 
        The current density (Acm2) to calculate the partial pressure at

    Returns: 
    ----------
    float
        The oxygen partial pressure (bar)
    """ 
    T_C        = stack_temperature - 273.15
    log_P_H2O = -2.1794 + 0.02953 * T_C - 9.1837e-5 * T_C**2 + 1.4454e-7 * T_C**3
    P_H2O = 10 ** log_P_H2O
    N = 0.291 * i / (stack_temperature ** 0.832)
    P_O2 = 0.21 * (P_air - P_drop / 2 - RH * P_H2O) *  ((1 + (air_excess_ratio - 1) / air_excess_ratio) / 2)/ np.exp(N)
    return P_O2

def calculate_P_H2(fuel_cell_stack, P_H2_input, stack_temperature, RH, i):
    """
    Calculate the partial pressure of hydroen at the fuel cell anode

    Parameters: 
    ----------
    P_H2_input: float 
        The input hydrogen pressure to the fuel cell (bar)
    stack_temperature: float 
        The fuel cell stack temperature (K)
    RH: float 
        The relative humidity of the oxygen stream (0-1)
    i: float 
        The current density (Acm2) to calculate the partial pressure at

    Returns: 
    ----------
    float
        The hydrogen partial pressure (bar)
    """
    T_C = stack_temperature - 273.15
    log_P_H2O = -2.1794 + 0.02953 * T_C - 9.1837e-5 * T_C**2 + 1.4454e-7 * T_C**3
    P_H2O = 10 ** log_P_H2O
    P_H2 = 0.5 * (P_H2_input / np.exp(1.653 * i / stack_temperature**1.334) - RH * P_H2O)
    return P_H2

def calculate_E_cell(fuel_cell_stack, stack_temperature, P_H2, P_O2):
    """
    Calculates the reversible Nernst Voltage of the cell

    Parameters:
    ----------
    stack_temperature: float 
        The fuel cell stack temperature (K)
    P_H2: float 
        The hydrogen partial pressure (bar)
    P_O2: float 
        The oxygen partial pressure (bar)

    Returns:
    ----------
    float 
        The reversible cell potential (V)
    """
    fuel_cell   = fuel_cell_stack.fuel_cell
    try:
        E_cell = 1.229 - 8.45e-4 * (stack_temperature - 298.15) + \
        fuel_cell.Universal_gas_constant*stack_temperature / (4 * fuel_cell.alpha * fuel_cell.Faraday_constant) * (np.log(P_H2) + 0.5 * np.log(P_O2))
    except: 
        return -10
    return E_cell

def calculate_activation_losses(fuel_cell_stack, stack_temperature, P_O2, i):
    """
    calculation of activation losses for both high temperature and low temperature fuel cells

    Parameters: 
    ----------
    stack_temperature: float 
        Stack Temperature (K)
    P_O2: float 
        Oxygen Partial Pressure (bar)
    i: float 
        Current density to evaluate activation losses (A/cm2)
    
    Returns: 
    ----------
    float: 
        Activation voltage loss (V)
    """
    fuel_cell = fuel_cell_stack.fuel_cell
    A_const   = fuel_cell.Universal_gas_constant * stack_temperature / (2 * fuel_cell.alpha * fuel_cell.Faraday_constant)
    i0        = fuel_cell.i0ref * fuel_cell.L_c * fuel_cell.a_c * (P_O2/fuel_cell.i0ref_P_ref) ** (fuel_cell.gamma) * \
        np.exp(-fuel_cell.E_C / (fuel_cell.Universal_gas_constant * stack_temperature) * (1 - (stack_temperature / fuel_cell.i0ref_T_ref)))
    eta_act = A_const * np.log(i/i0)
    return eta_act

def calculate_ohmic_losses_LT(fuel_cell_stack, stack_temperature, i):
    """
    Calculation of ohmic losses for low temperature fuel cells

    Parameters: 
    ----------
    stack_temperature: float 
        Stack Temperature (K)
    i: float 
        Current density to evaluate activation losses (A/cm2)
    
    Returns: 
    ----------
    float: 
        Ohmic voltage loss (V)
    """
    fuel_cell  = fuel_cell_stack.fuel_cell
    t_m        = fuel_cell.t_m 
    lambda_eff = fuel_cell.lambda_eff
    num        = 181.6 * (1 + 0.03 * i + 0.062 * (stack_temperature/303) ** 2 * i ** 2.5)
    denom      = (lambda_eff - 0.634 - 3 * i) * np.exp(4.18 * (stack_temperature - 303) / stack_temperature)
    rho        = num/denom 
    eta_ohmic  = (rho * t_m) * i
    return eta_ohmic 

def calculate_ohmic_losses_HT(fuel_cell_stack, stack_temperature, i):
    """
    Calculation of ohmic losses for high temperature fuel cells

    Parameters: 
    ----------
    stack_temperature: float 
        Stack Temperature (K)
    i: float 
        Current density to evaluate activation losses (A/cm2)
    
    Returns: 
    ----------
    float: 
        Ohmic voltage loss (V)
    """
    fuel_cell = fuel_cell_stack.fuel_cell
    t_m       = fuel_cell.t_m 
    c1        = fuel_cell.c1
    c2        = fuel_cell.c2
    s         = (stack_temperature - 373.15) / (100) * (c2 - c1) + c1
    rho       = 1/s 
    eta_ohmic = (rho * t_m) * i
    return eta_ohmic

def calculate_concentration_losses_LT(fuel_cell_stack, stack_temperature, P_O2, RH, air_excess_ratio, P_drop, i):
    """
    Calculation of concentration losses for low temperature fuel cells 

    Parameters:
    ----------
    stack_temperature: float 
        Stack Temperature (K)
    P_O2: float 
        Oxygen partial pressure (bar)
    RH: float 
        Relative humidity of the oxygen stream (0-1)
    air_excess_ratio: float 
        Oxygen excess ratio (above stoichometric)
    P_drop: float 
        Pressure drop across the fuel cell stack (bar)
    i: float 
        Current density to evaluate activation losses (A/cm2)
    
    Returns:
    ----------
    float: 
        Concentration voltage loss (V)
    """
    fuel_cell =  fuel_cell_stack.fuel_cell
    i_lim =  calculate_limiting_current_density_LT(fuel_cell_stack, stack_temperature, P_O2, RH, air_excess_ratio, P_drop, i)
    if i >= i_lim: 
        return 10
    else: 
        eta_conc = (1 + 1 / fuel_cell.alpha) * fuel_cell.Universal_gas_constant * stack_temperature / (2 * fuel_cell.Faraday_constant) * np.log(i_lim / (i_lim - i)) 
        return eta_conc
    
def calculate_limiting_current_density_LT(fuel_cell_stack, stack_temperature, P_O2, RH, air_excess_ratio, P_drop, i, **kwargs): 
    """
    Calculation of the limiting current density for LT-PEM

    Parameters:
    ----------
    stack_temperature: float 
        Stack Temperature (K)
    P_O2: float 
        Oxygen partial pressure (bar)
    RH: float 
        Relative humidity of the oxygen stream (0-1)
    air_excess_ratio: float 
        Oxygen excess ratio (above stoichometric)
    P_drop: float 
        Pressure drop across the fuel cell stack (bar)
    i: float 
        Current density to evaluate activation losses (A/cm2)
    
    Returns:
    -----------
    float: 
        limiting current density (A/cm^2)
    """
    fuel_cell        =  fuel_cell_stack.fuel_cell
    P_O2_ref_1_atm   = calculate_P_O2(fuel_cell_stack, 1, stack_temperature, RH, air_excess_ratio, P_drop, i)
    P_O2_ref_2_5_atm = calculate_P_O2(fuel_cell_stack, 2.5, stack_temperature, RH, air_excess_ratio, P_drop, i)
    i_lim            = (2.25 - 1.65) * (P_O2 - P_O2_ref_1_atm)**1 / (P_O2_ref_2_5_atm-P_O2_ref_1_atm)**1 + 1.65
    return fuel_cell.current_density_limit_multiplier * i_lim
    
def calculate_concentration_losses_HT(fuel_cell_stack, stack_temperature, P_O2, RH, air_excess_ratio, P_drop, i):
    """
    Calculation of concentration losses for high temperature fuel cells 

    Parameters:
    ----------
    stack_temperature: float 
        Stack Temperature (K)
    P_O2: float 
        Oxygen partial pressure (bar)
    RH: float 
        Relative humidity of the oxygen stream (0-1)
    air_excess_ratio: float 
        Oxygen excess ratio (above stoichometric)
    P_drop: float 
        Pressure drop across the fuel cell stack (bar)
    i: float 
        Current density to evaluate activation losses (A/cm2)
    
    Returns:
    -----------
    float: 
        Concentration voltage loss (V)
    """
    fuel_cell = fuel_cell_stack.fuel_cell
    i_lim = calculate_limiting_current_density_HT(fuel_cell_stack,stack_temperature, P_O2, RH, air_excess_ratio, P_drop, i)
    if i >= i_lim: 
        return 10
    else: 
        eta_conc = (1 + 1.8/fuel_cell.alpha) * fuel_cell.Universal_gas_constant * stack_temperature / (2* fuel_cell.Faraday_constant) * np.log(i_lim / (i_lim - i)) 
        return eta_conc

def calculate_limiting_current_density_HT(fuel_cell_stack, stack_temperature, P_O2, RH, air_excess_ratio, P_drop, i): 
    """
    Calculation of the limiting current density for HT-PEM

    Parameters:
    ----------
    stack_temperature: float 
        Stack Temperature (K)
    P_O2: float 
        Oxygen partial pressure (bar)
    RH: float 
        Relative humidity of the oxygen stream (0-1)
    air_excess_ratio: float 
        Oxygen excess ratio (above stoichometric)
    P_drop: float 
        Pressure drop across the fuel cell stack (bar)
    i: float 
        Current density to evaluate activation losses (A/cm2)
    
    Returns:
    -----------
    float: 
        limiting current density (A/cm^2)
    """
    fuel_cell      = fuel_cell_stack.fuel_cell
    P_O2_ref_1_atm = calculate_P_O2(fuel_cell_stack, 1, stack_temperature, RH, air_excess_ratio, P_drop, i)
    P_O2_ref_2_atm = calculate_P_O2(fuel_cell_stack,2, stack_temperature, RH, air_excess_ratio, P_drop, i)
    i_lim = (2.15 - 1.4) * (P_O2 - P_O2_ref_1_atm)**1 / (P_O2_ref_2_atm-P_O2_ref_1_atm)**1 + 1.4
    return fuel_cell.current_density_limit_multiplier * i_lim


def evaluate_max_gross_power(fuel_cell_stack,fuel_cell_conditions,t_idx):

    fuel_cell           = fuel_cell_stack.fuel_cell
    FC_air_p            = fuel_cell.rated_air_pressure 
    stack_temperature   = fuel_cell_conditions.fuel_cell.stack_temperature[t_idx,0]      
    RH                  = fuel_cell.oxygen_relative_humidity         
    air_excess_ratio    = fuel_cell.air_excess_ratio 
     
    P_O2 = calculate_P_O2(fuel_cell_stack,FC_air_p, stack_temperature, RH, air_excess_ratio, fuel_cell.rated_p_drop_fc, 0) 

    if fuel_cell.type == "LT":
        i_lim = calculate_limiting_current_density_LT(fuel_cell_stack,stack_temperature, P_O2, RH, air_excess_ratio, 0, i=0)
    elif fuel_cell.type == "HT":
        i_lim = calculate_limiting_current_density_HT(fuel_cell_stack,stack_temperature, P_O2, RH, air_excess_ratio, 0, i=0)  

    res = minimize_scalar(evaluate_power_func, args = (fuel_cell_stack,fuel_cell_conditions,t_idx), bounds = (0.2 * i_lim, 0.95 * i_lim))
    rated_current_density  = res.x 
    rated_power_density    = -res.fun 
     
    return rated_current_density, rated_power_density

def evaluate_power_func(i,fuel_cell_stack,fuel_cell_conditions,t_idx): 
    V_cell, V_loss = calculate_voltage(i,fuel_cell_stack,fuel_cell_conditions,t_idx)
    PD = -V_cell * i 
    return PD


# RCAIDE/Compoments/Wings/Control_Surfaces/Spoiler.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from .Control_Surface import Control_Surface 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Spoiler
# ----------------------------------------------------------------------------------------------------------------------
class Spoiler(Control_Surface):
    """
    A class representing a spoiler control surface for drag generation/lift reduction during descent
    and landing.

    Attributes
    ----------
    tag : str
        Unique identifier for the spoiler, defaults to 'spoiler' 

    Notes
    -----
    The spoiler is an upper surface device used as an airbrake to decrease the
    lift coefficient and increase drag. It inherits basic control surface functionality from 
    the Control_Surface class and adds specific attributes for spoiler operation. It is not modeled
    in the aerodynamics module, rather correlations from a NASA TR are used to predict aerodynamic
    coefficient modifications

    See Also
    --------
    RCAIDE.Library.Components.Wings.Control_Surfaces.Control_Surface
        Base class providing common control surface functionality
    RCAIDE.Library.Components.Wings.Control_Surfaces.Flap
        Trailing edge high-lift device
    """ 

    def __defaults__(self):
        """
        Sets default values for the spoiler attributes.
        
        Notes
        -----
        See Control_Surface.__defaults__ for additional inherited attributes.
        """
        self.tag            = 'spoiler' 
        
        pass
      


    


# RCAIDE/Methods/Mission/Segments/Single_Point/__init__.py
# 

"""
Collection of single-point analysis segments for aircraft mission simulation. This module 
provides segments for analyzing aircraft performance at specific flight conditions with 
various control parameters including speed, altitude, and throttle settings. Includes 
specialized segments for AVL (Athena Vortex Lattice) trimmed analysis.

See Also
--------
RCAIDE.Library.Mission.Segments.Climb
RCAIDE.Library.Mission.Segments.Cruise
RCAIDE.Library.Mission.Segments.Descent
RCAIDE.Library.Mission.Segments.Ground
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from . import Set_Speed_Set_Altitude
from . import Set_Speed_Set_Throttle 
from . import Set_Speed_Set_Altitude_AVL_Trimmed
from . import Set_Speed_Set_Altitude_No_Propulsion


# RCAIDE/Library/Methods/Aerodynamics/Common/Drag/spoiler_drag.py
# 
# Created:  Mar 2025, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import Units
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Spolier Drag 
# ---------------------------------------------------------------------------------------------------------------------- 
def spoiler_drag(state,settings,geometry):
    """Updates the aerodynamic performance of an aircraft given the influence of deflected spoilers
  
    Parameters
    ----------
    state : Data
        flight conditions of aircraft
        
    settings : dict
        aerodynamic settings
        
    geometry : Data
        aircraft geometry 

    References
    ----------
    [1]  Croom, Delwin R. Low-speed wind-tunnel investigation of various segments of flight spoilers
         as trailing-vortex-alleviation devices on a transport aircraft model. No. NASA-TN-D-8162. 1976.
      
    """    
    # unpack   
    parasite_total        = state.conditions.aerodynamics.coefficients.drag.parasite.total            
    induced_total         = state.conditions.aerodynamics.coefficients.drag.induced.total            
    compressibility_total = state.conditions.aerodynamics.coefficients.drag.compressible.total     
    miscellaneous_drag    = state.conditions.aerodynamics.coefficients.drag.miscellaneous.total
    cooling_drag          = state.conditions.aerodynamics.coefficients.drag.cooling.total  

    # untrimmed drag 
    drag  =  parasite_total + induced_total  + compressibility_total + miscellaneous_drag + cooling_drag 
    
    spoiler_drag_coef =  np.zeros_like(drag)
    for wing in geometry.wings: 
        for cs in wing.control_surfaces:
            if type(cs) == RCAIDE.Library.Components.Wings.Control_Surfaces.Spoiler:
                spoiler_drag_coef += drag * (0.0011 * (cs.deflection / Units.degrees))  
                state.conditions.aerodynamics.coefficients.lift.total  +=  -0.0075 *(cs.deflection / Units.degrees)  
                state.conditions.static_stability.coefficients.M       +=  0.0053 *(cs.deflection / Units.degrees)  
    
    state.conditions.aerodynamics.coefficients.drag.spoiler.total =  spoiler_drag_coef 
    return  



# RCAIDE/Library/Methods/Mass_Properties/Weight_Buildups/Conventional/BWB/FLOPS/__init__.py
# 

"""
Methods for computing component weights specific to Blended Wing Body (BWB) aircraft configurations. 
This module provides specialized weight estimation techniques that account for the unique structural 
and geometric characteristics of BWB designs.

See Also
--------
RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Common
    Shared component weight methods
RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport
    Transport aircraft weight methods
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_operating_empty_weight   import compute_operating_empty_weight
from .compute_aft_centerbody_weight    import compute_aft_centerbody_weight
from .compute_cabin_weight             import compute_cabin_weight


# RCAIDE/Library/Methods/Powertrain/Converters/Ram/__init__.py

"""
This module provides functionality for modeling ram air compression in powertrains. It includes methods for computing 
ram compression performance and appending ram conditions to simulation results.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .append_ram_conditions   import append_ram_conditions                         
from .compute_ram_performance import compute_ram_performance


# RCAIDE/Methods/Noise/Correlation_Buildup/Airframe/clean_wing_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
import numpy as np
from RCAIDE.Framework.Core import Units

# ----------------------------------------------------------------------------------------------------------------------  
# main and nose landing gear noise
# ----------------------------------------------------------------------------------------------------------------------  
def landing_gear_noise(D,H,wheels,M,velocity,phi,theta,distance,frequency):
    """This calculates the Landing gear 1/3 octave band sound pressure level and overall sound pressure level
    for a tyre diameter D, a strut length H and WHEELS number of  wheels per unit.

    Assumptions:
        Correlation based.

    Source:
       Fink, Martin R. "Noise component method for airframe noise." Journal of aircraft 16.10 (1979): 659-665.


    Inputs:
        D         - Landing gear tyre diameter                                      
        H         - Lading gear strut length                                        
        wheels    - Number of wheels per unit                                      [-]
        M         - Mach number                                                    [-]
        velocity  - Aircraft speed                                                  
        phi       - Azimuthal angle                                                [rad]
        theta     - Polar angle                                                    [rad]
        distance  - Distance from airplane to observer, evaluated at retarded time [ft]
        frequemcy - Frequency array                                                [Hz] 

    Outputs: One Third Octave Band SPL [dB]
        SPL           - Sound Pressure Level of the landing gear         [dB]
        OASPL         - Overall Sound Pressure Level of the landing gear [dB] 
 
    """  
     
    velocity_kts = velocity/Units.knots

    if (wheels==1 or wheels==2):
        G1 =  130
        G2 = 10 *np.log10(4.5* ((frequency*D/velocity)**2) * (12.5 + ((frequency*D/velocity)**2) )**(-2.25) )   
    else:
        G1 = 123
        G2 = 10 *np.log10(0.3* ((frequency*D/velocity)**2) * (1  + 0.25*((frequency*D/velocity)**2) )**(-1.5) )    
 
    SPL   = 60.*np.log10(velocity_kts/194.0)+20.*np.log10(D/distance)+ G1 + G2 

    return SPL



# RCAIDE/Library/Methods/Powertrain/Modulators/Electronic_Speed_Controller/append_motor_conditions.py
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_esc_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_esc_conditions(esc,segment,energy_conditions): 
    """
    Initializes the Electronic Speed Controller (ESC) condition containers for tracking 
    electrical state variables. Sets up basic input/output conditions and throttle settings 
    for ESC performance analysis.

    Parameters
    ----------
    esc : RCAIDE.Library.Components.Energy.Modulators.Electronic_Speed_Controller
        The electronic speed controller component
            - tag : str
                Identifier for the ESC
    segment : RCAIDE.Framework.Mission.Segments.Segment
        The mission segment being analyzed
            - state : State
                Contains the flight condition state variables
                    - ones_row : function
                        Returns array of ones with specified size
    propulsor_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Container for propulsor-specific conditions

    Returns
    -------
    None

    Notes
    -----
    Creates and initializes the following state variables:
        - inputs : Conditions
            Container for input electrical parameters
        - outputs : Conditions
            Container for output electrical parameters
        - throttle : float
            Power modulation setting from 0 to 1
    """
    
    ones_row    = segment.state.ones_row 
    energy_conditions.modulators[esc.tag]                  = Conditions()
    energy_conditions.modulators[esc.tag].inputs           = Conditions()
    energy_conditions.modulators[esc.tag].outputs          = Conditions()
    energy_conditions.modulators[esc.tag].throttle         = 0. * ones_row(1)  
    energy_conditions.modulators[esc.tag].outputs.voltage  = 0. * ones_row(1)  
    energy_conditions.modulators[esc.tag].inputs.voltage   = esc.bus_voltage * ones_row(1)   
    return 



# RCAIDE/Library/Plots/Performance/Stability/plot_longitudinal_stability.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_longitudinal_stability(results,
                             save_figure = False,
                             show_legend = True,
                             save_filename = "Longitudinal_Stability",
                             file_type = ".png",
                             width = 11, height = 7):
    """
    Creates a multi-panel visualization of longitudinal stability characteristics.

    Parameters
    ----------
    results : Results
        RCAIDE results data structure containing:
            - segments[i].conditions.frames.inertial.time[:,0]
                Time history for each segment
            - segments[i].conditions.aerodynamics.angles.theta[:,0]
                Pitch angle history
            - segments[i].conditions.aerodynamics.angles.alpha[:,0]
                Angle of attack history
            - segments[i].conditions.control_surfaces.elevator.deflection[:,0]
                Elevator deflection history
            - segments[i].tag
                Name/identifier of each segment
            
    save_figure : bool, optional
        Flag for saving the figure (default: False)
        
    show_legend : bool, optional
        Flag to display segment legend (default: True)
        
    save_filename : str, optional
        Name of file for saved figure (default: "Longitudinal_Stability")
        
    file_type : str, optional
        File extension for saved figure (default: ".png")
        
    width : float, optional
        Figure width in inches (default: 11)
        
    height : float, optional
        Figure height in inches (default: 7)

    Returns
    -------
    fig : matplotlib.figure.Figure

    Notes
    -----
    Creates visualization showing:
        * Pitch attitude response
        * Aerodynamic angle evolution
        * Longitudinal control inputs
        * Trajectory angle history
        * Time history for each segment
    
    **Major Assumptions**
        * Angles are in degrees
        * Time is in minutes
        * Positive deflections follow right-hand rule
   
    **Definitions**
    
    'Pitch Angle'
        Nose-up/down attitude relative to horizon
    'Angle of Attack'
        Angle between velocity vector and body reference line
    'Flight Path Angle'
        Angle between velocity vector and horizon
    'Elevator Deflection'
        Pitch control surface position
    
    See Also
    --------
    RCAIDE.Library.Plots.Stability.plot_lateral_stability : Lateral-directional stability analysis
    RCAIDE.Library.Plots.Stability.plot_flight_forces_and_moments : Force/moment visualization
    """  

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     
     
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    

    axis_1 = plt.subplot(3,2,1)    
    axis_2 = plt.subplot(3,2,2) 
    axis_3 = plt.subplot(3,2,3)  
    axis_4 = plt.subplot(3,2,4)  
    axis_5 = plt.subplot(3,2,5)
    axis_6 = plt.subplot(3,2,6) 
    
    for i in range(len(results.segments)): 
        time       = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min
        c_m        = results.segments[i].conditions.static_stability.coefficients.M[:,0]   
        SM         = results.segments[i].conditions.static_stability.static_margin[:,0]  
        delta_e    = results.segments[i].conditions.control_surfaces.elevator.deflection[:,0] / Units.deg
        CM_delta_e = results.segments[i].conditions.static_stability.derivatives.CM_delta_e[:,0]
        Cm_alpha   = results.segments[i].conditions.static_stability.derivatives.CM_alpha[:,0]
        CL_alpha   = results.segments[i].conditions.static_stability.derivatives.Clift_alpha[:,0] 
          
        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ')  
        
        axis_1.plot(time, c_m, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = segment_name) 
        axis_1.set_ylabel(r'$C_M$')
        axis_1.set_ylim([-1, 1])  
        set_axes(axis_1) 

        axis_2.plot(time, Cm_alpha, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width) 
        axis_2.set_ylabel(r'$C_M\alpha$')
        axis_2.set_ylim([-1, 1]) 
        set_axes(axis_2) 
        
        axis_3.plot(time,SM , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_3.set_xlabel('Time (mins)')
        axis_3.set_ylabel(r'Static Margin (%)')
        set_axes(axis_3)  

        axis_4.plot(time,delta_e , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_4.set_xlabel('Time (mins)')
        axis_4.set_ylabel(r'Elevator Defl.n') 
        axis_4.set_ylim([-15, 15]) 
        set_axes(axis_4) 
        
        axis_5.plot(time,CM_delta_e , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_5.set_xlabel('Time (mins)')
        axis_5.set_ylabel(r'$C_M\delta_e$')
        axis_5.set_ylim([-1, 1]) 
        set_axes(axis_5)
        
        axis_6.plot(time,CL_alpha, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_6.set_xlabel('Time (mins)')
        axis_6.set_ylabel(r'$C_L\alpha$')
        axis_6.set_ylim([-1, 10]) 
        set_axes(axis_6)    
        
    if show_legend:
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})     
          
    # Adjusting the sub-plots for legend
    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'Stability Coefficients'      
    fig.suptitle(title_text)
 
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return fig 
 


# RCAIDE/Library/Missions/Segments/Cruise/Curved_Constant_Radius_Constant_Speed_Constant_Altitude/initialize_conditions.py
# 
# 
# Created:  September 2024, A. Molloy + M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# Package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """
    Initializes conditions for constant radius curved flight at fixed altitude and speed.

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - altitude : float
                Cruise altitude [m]
            - air_speed : float
                True airspeed to maintain [m/s]
            - true_course : float
                Initial true course angle [deg]
            - turn_angle : float
                Total turn angle to execute [deg]
                Positive for right turn, negative for left turn
            - turn_radius : float
                Radius of the turn [m]
            - sideslip_angle : float
                Aircraft sideslip angle [rad]
            - state:
                numerics.dimensionless.control_points : array
                    Discretization points [-]
                conditions : Data
                    State conditions container
                initials : Data, optional
                    Initial conditions from previous segment
        
    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.frames.inertial.velocity_vector [m/s]
            - conditions.frames.inertial.position_vector [m]
            - conditions.frames.body.velocity_vector [m/s]
            - conditions.freestream.altitude [m]
            - conditions.frames.inertial.time [s]
            - conditions.frames.planet.true_heading [rad]
            - conditions.frames.planet.true_course [rad]
    
    Notes
    -----
    This function sets up the initial conditions for a coordinated turn segment with 
    constant radius, constant speed, and constant altitude. The turn direction and 
    magnitude are specified by the turn angle.

    **Calculation Process**
    1. Check initial conditions
    2. Calculate turn rate from speed and radius:
        ω = V/R where:
            - V is true airspeed
            - R is turn radius
    3. Calculate time required for turn:
        t = |θ|/ω where:
            - θ is turn angle
            - ω is turn rate
    4. Discretize time points
    5. Calculate true course progression
    6. Decompose velocity into inertial and body components

    **Major Assumptions**
        * Coordinated turn (true course aligned with true heading)
        * Constant true airspeed
        * Constant altitude
        * Constant turn radius
        * Small angle approximations
        * Quasi-steady flight

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """        
    
    # unpack 
    alt               = segment.altitude
    air_speed         = segment.air_speed       
    beta              = segment.sideslip_angle
    radius            = segment.turn_radius
    start_true_course = segment.true_course
    arc_sector        = segment.turn_angle
    conditions        = segment.state.conditions 

    # check for initial velocity
    if air_speed is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # check for turn radius
    if radius is None:
        if not segment.state.initials: raise AttributeError('radius not set')
        radius = 0.1 # minimum radius so as to approximate a near instantaneous curve
    
    # check for turn angle
    if arc_sector is None:
        if not segment.state.initials: raise AttributeError('turn angle not set')
        arc_sector = 0.0 # aircraft does not turn    

    # dimensionalize time
    v_body_x    = np.cos(beta)*air_speed # x-velocity in the body frame. 
    v_body_y    = np.sin(beta)*air_speed # y-velocity in the body frame
    t_initial   = conditions.frames.inertial.time[0,0]
    omega       = v_body_x / radius
    t_final     = abs(arc_sector) / omega + t_initial  # Time to complete the turn
    t_nondim    = segment.state.numerics.dimensionless.control_points
    time        = t_nondim * (t_final-t_initial) + t_initial
    
    true_course_control_points = start_true_course + t_nondim * arc_sector
    
    v_inertial_x = air_speed * np.cos(true_course_control_points)
    v_inertial_y = air_speed * np.sin(true_course_control_points)
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_inertial_x[:,0]
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_inertial_y[:,0]
    segment.state.conditions.frames.body.velocity_vector[:,0]     = v_body_x
    segment.state.conditions.frames.body.velocity_vector[:,1]     = v_body_y
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]
    segment.state.conditions.frames.planet.true_heading[:,0]      = true_course_control_points[:,0]
    segment.state.conditions.frames.planet.true_course[:,0]       = true_course_control_points[:,0]


# RCAIDE/Library/Methods/Powertrain/Propulsors/Turbofan/append_turbofan_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports  
import RCAIDE
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_propulsor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_turbofan_conditions(propulsor, segment, energy_conditions, noise_conditions):
    """
    Initializes turbofan operating conditions for a mission segment.
    
    Parameters
    ----------
    propulsor : RCAIDE.Library.Components.Propulsors.Turbofan
        Turbofan propulsor component with the following attributes:
            - tag : str
                Identifier for the turbofan
            - items : dict
                Dictionary of subcomponents
    segment : RCAIDE.Framework.Mission.Segments.Segment
        Mission segment with the following attributes:
            - state : Data
                Segment state
                    - ones_row : function
                        Function to create array of ones with specified length
    energy_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Energy conditions container where turbofan conditions will be stored
    noise_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Noise conditions container where turbofan noise conditions will be stored
    
    Returns
    -------
    None
        Results are stored in energy_conditions.propulsors[propulsor.tag] and
        noise_conditions.propulsors[propulsor.tag]
    
    Notes
    -----
    This function initializes the necessary data structures for storing turbofan
    operating conditions during a mission segment. It creates zero-filled arrays
    for various performance parameters and recursively calls the append_operating_conditions
    method for each subcomponent of the turbofan.
    
    The function initializes the following parameters in energy_conditions:
        * throttle
        * commanded_thrust_vector_angle
        * thrust
        * power
        * moment
        * fuel_flow_rate
        * inputs and outputs containers
    
    It also creates the following containers in noise_conditions:
        * core_nozzle
        * fan_nozzle
        * fan
    
    **Major Assumptions**
        * All arrays are initialized with zeros
        * Each component has an append_operating_conditions method
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbofan.compute_turbofan_performance
    RCAIDE.Library.Methods.Powertrain.Propulsors.Turbofan.compute_thurst
    """
    # unpack 
    ones_row          = segment.state.ones_row 
    
    # add propulsor conditions          
    energy_conditions.propulsors[propulsor.tag]                               = Conditions()  
    energy_conditions.propulsors[propulsor.tag].throttle                      = 0. * ones_row(1)      
    energy_conditions.propulsors[propulsor.tag].commanded_thrust_vector_angle = 0. * ones_row(1)  
    energy_conditions.propulsors[propulsor.tag].thrust                        = 0. * ones_row(3) 
    energy_conditions.propulsors[propulsor.tag].power                         = 0. * ones_row(1) 
    energy_conditions.propulsors[propulsor.tag].moment                        = 0. * ones_row(3) 
    energy_conditions.propulsors[propulsor.tag].fuel_flow_rate                = 0. * ones_row(1)
    energy_conditions.propulsors[propulsor.tag].inputs                        = Conditions()
    energy_conditions.propulsors[propulsor.tag].outputs                       = Conditions() 
    noise_conditions.propulsors[propulsor.tag]                                = Conditions()  
    noise_conditions.propulsors[propulsor.tag].core_nozzle                    = Conditions() 
    noise_conditions.propulsors[propulsor.tag].fan_nozzle                     = Conditions() 
    noise_conditions.propulsors[propulsor.tag].fan                            = Conditions()
 
    for tag, item in  propulsor.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,energy_conditions) 
            for sub_tag, sub_item in  item.items(): 
                if issubclass(type(sub_item), RCAIDE.Library.Components.Component): 
                    sub_item.append_operating_conditions(segment,energy_conditions)    
    return 


# RCAIDE/Methods/Powertrain/Sources/Batteries/Lithium_Ion_LFP/update_lfp_cell_age.py
# 
# 
# Created: Nov 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import numpy as np  
 
# ----------------------------------------------------------------------------------------------------------------------
# update_lfp_cell_age
# ----------------------------------------------------------------------------------------------------------------------  
def update_lfp_cell_age(battery_module, segment, battery_conditions, increment_battery_age_by_one_day):  
    """
    Updates the aging model for a 26650 A123 LFP cell.
    
    Parameters
    ----------
    battery_module : BatteryModule
        The battery module containing LFP cells
    segment : Segment
        The mission segment in which the battery is operating
    battery_conditions : Conditions
        Object containing battery state with the following attributes:
            - cell.state_of_charge : numpy.ndarray
                State of charge of the cell [unitless, 0-1]
            - cell.current : numpy.ndarray
                Battery cell current [A]
            - cell.cycle_in_day : int
                Number of cycles the battery has undergone [days]
            - cell.charge_throughput : numpy.ndarray
                Cumulative charge throughput [Ah]
            - cell.temperature : numpy.ndarray
                Battery cell temperature [K]
            - cell.capacity_fade_factor : float
                Factor representing capacity degradation [unitless, 0-1]
    increment_battery_age_by_one_day : bool
        Flag to increment the battery age by one day
    
    Returns
    -------
    None
    
    Notes
    -----
    This function implements a semi-empirical aging model for LFP cells based on
    research by Nájera et al. The model accounts for capacity fade due to:
        1. Cycling effects (charge throughput)
        2. Calendar aging (time)
    
    The model considers the effects of:
        - Temperature
        - C-rate
        - State of charge
        - Charge throughput
        - Time (days)
    
    References
    ----------
    [1] Nájera, J., J.R. Arribas, R.M. De Castro, and C.S. Núñez. "Semi-Empirical Ageing Model for LFP and NMC Li-Ion Battery Chemistries." Journal of Energy Storage 72 (November 2023): 108016. https://doi.org/10.1016/j.est.2023.108016.
    
    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Sources.Battery_Modules.Lithium_Ion_LFP
    RCAIDE.Library.Methods.Powertrain.Sources.Batteries.Lithium_Ion_LFP.compute_lfp_cell_performance
    """
    SOC                = battery_conditions.cell.state_of_charge
    I                  = battery_conditions.cell.current
    t                  = battery_conditions.cell.cycle_in_day         
    charge_thougput    = battery_conditions.cell.charge_throughput
    Temp               = (battery_conditions.cell.temperature) 
    C_rate             = np.sqrt(np.mean(I**2)) /battery_module.cell.nominal_capacity
    
    # Semi Emperical aging model  
    E_fade_factor = 1-(((Temp**2*2.0916e-8)+(-1.2179e-5*Temp)+0.0018)*np.exp(((-1.7082e-6*Temp)+0.0556)*C_rate) \
                        * charge_thougput + (5.9808e6) * np.exp(0.68989*SOC) * np.exp(-6.4647e3/Temp) * t**(0.5))


    battery_conditions.cell.capacity_fade_factor     = np.minimum(E_fade_factor[-1],battery_conditions.cell.capacity_fade_factor)
    
    if increment_battery_age_by_one_day:
        battery_conditions.cell.cycle_in_day += 1 # update battery age by one day 
    
    return 



## @ingroup Library-Plots-Energy
# RCAIDE/Library/Plots/Energy/plot_altitude_sfc_weight.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
## @ingroup Library-Plots-Performance-Energy-Fuel
def plot_altitude_sfc_weight(results,
                             save_figure = False,
                             show_legend = True,
                             save_filename = "Weight_Fuel_Consumption" ,
                             file_type = ".png",
                             width = 11, height = 7):
    """
    Creates a four-panel plot showing throttle settings, vehicle weight, specific fuel consumption (SFC), 
    and fuel consumption rate over time.

    Parameters
    ----------
    results : Results
        RCAIDE results structure containing segment data
        
    save_figure : bool, optional
        Flag for saving the figure (default: False)
        
    show_legend : bool, optional
        Flag for displaying plot legend (default: True)
        
    save_filename : str, optional
        Name of file for saved figure (default: "Altitude_SFC_Weight")
        
    file_type : str, optional
        File extension for saved figure (default: ".png")
        
    width : float, optional
        Figure width in inches (default: 11)
        
    height : float, optional
        Figure height in inches (default: 7)

    Returns
    -------
    fig : matplotlib.figure.Figure
        Figure handle containing the generated plots

    Notes
    -----
    The function creates a 2x2 subplot containing:
        2. Vehicle weight vs time
        1. Fuel consumption vs time
        3. Specific fuel consumption vs time
        4. Fuel consumption rate vs time
    
    Each segment is plotted with a different color from the inferno colormap.
  
    **Definitions**
    
    'SFC'
        Specific Fuel Consumption - measure of the fuel efficiency of an engine design 
        with respect to thrust output
    """
 
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))      
     
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height) 
    axis_1 = plt.subplot(2,2,1)
    axis_2 = plt.subplot(2,2,2)
    axis_3 = plt.subplot(2,2,3)
    axis_4 = plt.subplot(2,2,4)
    
    for i in range(len(results.segments)): 
        time      = results.segments[i].conditions.frames.inertial.time[:, 0] / Units.min
        Weight    = results.segments[i].conditions.weights.total_mass[:, 0] * 9.81   
        mdot      = results.segments[i].conditions.weights.vehicle_mass_rate[:, 0]
        thrust    = results.segments[i].conditions.frames.body.thrust_force_vector[:, 0]
        thrust    = results.segments[i].conditions.frames.body.thrust_force_vector[:, 0]
        fuel_mass = results.segments[i].conditions.energy.cumulative_fuel_consumption[:, 0]
        sfc       = (mdot / Units.lb) / (thrust / Units.lbf) * Units.hr
        

        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ')       
        axis_1.plot(time, Weight/1000 , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = segment_name )  
        axis_1.set_ylabel(r'Weight (kN)')  
        set_axes(axis_1) 

        axis_2.plot(time, fuel_mass, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_2.set_xlabel('Time (mins)')
        axis_2.set_ylabel(r'Fuel Consumption (kg)')
        set_axes(axis_2)
        

        axis_3.plot(time, sfc, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_3.set_xlabel('Time (mins)')
        axis_3.set_ylabel(r'SFC (lb/lbf-hr)')
        set_axes(axis_3) 

        axis_4.plot(time, mdot, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_4.set_xlabel('Time (mins)')
        axis_4.set_ylabel(r'Fuel Rate (kg/s)')
        set_axes(axis_4)         
        
    
    if show_legend:
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4)    
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'Weight and Fuel Consumption'      
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return fig 


# RCAIDE/Library/Methods/Powertrain/Modulators/Electronic_Speed_Controller/__init__.py
# 

"""
Collection of methods for analyzing Electronic Speed Controller (ESC) performance in electric 
propulsion systems. These devices, ESCs, handle voltage and current modulation for electric motors, 
including throttle response, power conversion, and condition tracking. The module provides 
functionality for computing ESC input/output characteristics and managing electrical state 
variables throughout mission analysis.

See Also
--------
RCAIDE.Library.Components.Powertrain.Converters.DC_Motor
RCAIDE.Library.Components.Powertrain.Modulators.Electronic_Speed_Controller
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .append_esc_conditions  import  append_esc_conditions
from .compute_esc_performance import compute_voltage_out_from_throttle
from .compute_esc_performance import compute_current_in_from_throttle 



# RCAIDE/Library/Methods/Powertrain/Converters/Ducted_Fan/__init__.py

# Created:  Jan 2025, M. Clarke

"""
Ducted Fan Methods Package

This module contains methods for analyzing and designing ducted fan propulsion systems. The methods support both low and medium-fidelity analysis.

The module provides functionality for:
    - Geometry generation and manipulation
    - Performance analysis and efficiency calculations  
    - DFDC case setup and execution
    - Results processing and data handling

See Also
--------
RCAIDE.Library.Components.Powertrain.Converters.Ducted_Fan
RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.compute_ducted_fan_performance
RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.design_ducted_fan
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .compute_ducted_fan_performance     import compute_ducted_fan_performance 
from .append_ducted_fan_conditions       import append_ducted_fan_conditions 
from .design_ducted_fan                  import design_ducted_fan
from .                                   import Performance


# RCAIDE/Library/Missions/Common/Unpack_Unknowns/energy.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  Unpack Unknowns
# ---------------------------------------------------------------------------------------------------------------------- 
def unknowns(segment):  
    ACV_T      =  segment.assigned_control_variables.throttle
    ACV_TA     =  segment.assigned_control_variables.thrust_vector_angle
    ACV_RBPC   =  segment.assigned_control_variables.rotor_blade_pitch_command
    
    for network in segment.analyses.energy.vehicle.networks: 
        if 'throttle' in segment: 
            for propulsor in network.propulsors: 
                segment.state.conditions.energy.propulsors[propulsor.tag].throttle[:,0] = segment.throttle
            
        if ACV_T.active: 
            for i in range(len(ACV_T.assigned_propulsors)): 
                propulsor_group = ACV_T.assigned_propulsors[i]
                for propulsor_name in propulsor_group:  
                    segment.state.conditions.energy.propulsors[propulsor_name].throttle = segment.state.unknowns["throttle_" + str(i)]  
    
       ## Thrust Vector Control 
        #if ACV_TA.active:                
            #for i in range(len(ACV_TA.assigned_propulsors)): 
                #propulsor_group = ACV_TA.assigned_propulsors[i]
                #for propulsor_name in propulsor_group:  
                    #segment.state.conditions.energy.propulsors[propulsor_name].commanded_thrust_vector_angle = segment.state.unknowns["thrust_vector_" + str(i)]
                    
             
        ## Thrust Vector Control 
        #if ACV_RBPC.active:                
            #for i in range(len(ACV_RBPC.assigned_propulsors)): 
                #propulsor_group = ACV_RBPC.assigned_propulsors[i] 
                #for propulsor_name in propulsor_group:
                    #propulsor = network.propulsors[propulsor_name]
                    #rotor     = propulsor.rotor 
                    #segment.state.conditions.energy.propulsors[propulsor_name][rotor.tag].blade_pitch_command = segment.state.unknowns["blade_pitch_command_" + str(i)]                    
    return 
     
 
    



# RCAIDE/Methods/Powertrain/Sources/Croygenic_Tanks/Common/__init__.py
# 

"""
This module provides functionality for modeling cryogenic tank systems in powertrains.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Sources
RCAIDE.Library.Methods.Powertrain.Sources.Fuel_Tanks
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from .append_cryogenic_tank_conditions   import append_cryogenic_tank_conditions


# RCAIDE/Library/Missions/Common/Pre_Process/energy.py
# 
# 
# Created:  Jul 2023, M. Clarke
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE
# ----------------------------------------------------------------------------------------------------------------------  
import RCAIDE
 
# ----------------------------------------------------------------------------------------------------------------------
#  energy
# ----------------------------------------------------------------------------------------------------------------------  
def energy(mission):
    """ Pre-processes energy network by appending all unknowns and residuals             
    """       
    for segment in mission.segments: 
        for network in segment.analyses.energy.vehicle.networks: 
            if type(network) == RCAIDE.Framework.Networks.Hybrid:
                if segment.hybrid_power_split_ratio == None:
                    raise AssertionError('Hybridization power split ratio not set! Specify in mission segment') 
                if segment.battery_fuel_cell_power_split_ratio == None:
                    raise AssertionError('Battery/Fuel cell power split ratio not set! Specify in mission segment')                 
            elif type(network) == RCAIDE.Framework.Networks.Fuel: 
                if segment.hybrid_power_split_ratio == None:                
                    segment.hybrid_power_split_ratio = 0.0
                    segment.battery_fuel_cell_power_split_ratio = 0.0
            elif type(network) == RCAIDE.Framework.Networks.Electric: 
                if segment.hybrid_power_split_ratio == None:                
                    segment.hybrid_power_split_ratio = 1.0  
                    segment.battery_fuel_cell_power_split_ratio = 1.0
            elif type(network) == RCAIDE.Framework.Networks.Fuel_Cell: 
                if segment.hybrid_power_split_ratio == None:                
                    segment.hybrid_power_split_ratio = 1.0  
                    segment.battery_fuel_cell_power_split_ratio = 0.0
            segment.state.conditions.energy.hybrid_power_split_ratio            = segment.hybrid_power_split_ratio * segment.state.ones_row(1)  
            segment.state.conditions.energy.battery_fuel_cell_power_split_ratio = segment.battery_fuel_cell_power_split_ratio * segment.state.ones_row(1)                    
            network.add_unknowns_and_residuals_to_segment(segment) 
    return 


# RCAIDE/Library/Methods/Powertrain/Propulsors/Common/
# 
# Created:  Jan 2025, M. Clarke  


# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports
import RCAIDE   
from RCAIDE.Framework.Mission.Common import Results, Residuals
from RCAIDE.Library.Mission.Common.Update.orientations import orientations

# Python package imports
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  Operating Test Conditions Set-up
# ---------------------------------------------------------------------------------------------------------------------- 
def setup_operating_conditions(component, altitude=0, velocity_range=np.array([10]), angle_of_attack=0):
    """
    Sets up operating conditions for single component analysis.
    
    Parameters
    ----------
    component : RCAIDE.Library.Components.Component
        Component to set up operating conditions for
            - working_fluid : Data
                Working fluid properties object (will be set by this function)
    altitude : float, optional
        Altitude for analysis [m]
        Default: 0 (sea level)
    velocity_range : numpy.ndarray, optional
        Array of velocities to analyze [m/s]
        Default: np.array([10])
    angle_of_attack : float, optional
        Angle of attack for analysis [deg]Default: 0
    
    Returns
    -------
    state : RCAIDE.Framework.Mission.Common.State
        State object containing:
            - conditions : Data
                Flight conditions
                    - freestream : Data
                        Freestream properties
                            - altitude : numpy.ndarray
                                Altitude [m]
                            - mach_number : numpy.ndarray
                                Mach number
                            - pressure : numpy.ndarray
                                Atmospheric pressure [Pa]
                            - temperature : numpy.ndarray
                                Atmospheric temperature [K]
                            - density : numpy.ndarray
                                Air density [kg/m³]
                            - dynamic_viscosity : numpy.ndarray
                                Air dynamic viscosity [kg/(m·s)]
                            - gravity : numpy.ndarray
                                Gravitational acceleration [m/s²]
                            - isentropic_expansion_factor : numpy.ndarray
                                Ratio of specific heats (gamma)
                            - Cp : numpy.ndarray
                                Specific heat at constant pressure [J/(kg·K)]
                            - R : numpy.ndarray
                                Gas constant [J/(kg·K)]
                            - speed_of_sound : numpy.ndarray
                                Speed of sound [m/s]
                            - velocity : numpy.ndarray
                                Freestream velocity [m/s]
                    - frames : Data
                        Reference frames
                            - body : Data
                                Body-fixed frame
                                    - inertial_rotations : numpy.ndarray
                                        Rotation angles [rad]
                            - inertial : Data
                                Inertial frame
                                    - velocity_vector : numpy.ndarray
                                        Velocity vector [m/s]
    
    Notes
    -----
    This function creates a standardized set of operating conditions for component analysis.
    It sets up atmospheric conditions based on the US Standard Atmosphere 1976 model and
    initializes all necessary parameters for component performance evaluation.
    
    The function:
        1. Sets up Earth as the planet and air as the working fluid
        2. Computes atmospheric properties at the specified altitude
        3. Creates a conditions data structure with all necessary parameters
        4. Sets up reference frames and orientations
        5. Appends component-specific operating conditions
    
    **Major Assumptions**
        * US Standard Atmosphere 1976
        * Earth gravity
        * Air as working fluid
    
    See Also
    --------
    RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976
    RCAIDE.Library.Mission.Common.Update.orientations
    """
    
    planet                                            = RCAIDE.Library.Attributes.Planets.Earth()
    working_fluid                                     = RCAIDE.Library.Attributes.Gases.Air() 
    
    # append working fluid properties 
    component.working_fluid                           = working_fluid     
    
    atmosphere_sls                                    = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data                                         = atmosphere_sls.compute_values(altitude,0.0) 
    p                                                 = atmo_data.pressure          
    T                                                 = atmo_data.temperature       
    rho                                               = atmo_data.density          
    a                                                 = atmo_data.speed_of_sound    
    mu                                                = atmo_data.dynamic_viscosity 
                                                      
    conditions                                        = Results() 
    conditions.freestream.altitude                    = np.atleast_2d(altitude)
    conditions.freestream.mach_number                 = np.atleast_2d(velocity_range/a)
    conditions.freestream.pressure                    = np.atleast_2d(p)
    conditions.freestream.temperature                 = np.atleast_2d(T)
    conditions.freestream.density                     = np.atleast_2d(rho)
    conditions.freestream.dynamic_viscosity           = np.atleast_2d(mu)
    conditions.freestream.gravity                     = np.atleast_2d(planet.sea_level_gravity)
    conditions.freestream.isentropic_expansion_factor = np.atleast_2d(working_fluid.compute_gamma(T,p))
    conditions.freestream.Cp                          = np.atleast_2d(working_fluid.compute_cp(T,p))
    conditions.freestream.R                           = np.atleast_2d(working_fluid.gas_specific_constant)
    conditions.freestream.speed_of_sound              = np.atleast_2d(a)

    num_ctrl_pts      = len(velocity_range)    
    conditions._size  = num_ctrl_pts
    conditions.expand_rows(num_ctrl_pts)
     
    conditions.freestream.velocity                    = np.atleast_2d(velocity_range) 
    conditions.frames.body.inertial_rotations[:, 1]   = angle_of_attack
    conditions.frames.inertial.velocity_vector[:, 0]  = np.atleast_2d(velocity_range)

    # setup conditions   
    segment                                          = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions                         = conditions    
    orientations(segment) 
    segment.state.residuals.network                  = Residuals()
    
    # append component-specific operating conditions 
    component.append_operating_conditions(segment,segment.state.conditions.energy,segment.state.conditions.noise)    
    segment.state.conditions.expand_rows(num_ctrl_pts)              
    return segment.state
 
    
    


# RCAIDE/Library/Methods/Propulsor/Ducted_Fan_Propulsor/purge_files.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------

import os
# ---------------------------------------------------------------------------------------------------------------------- 
# Purge Files  
# ----------------------------------------------------------------------------------------------------------------------   
def purge_files(filenames_array, directory=''):
    """
    Removes specified files from a directory to prevent conflicts.
    
    Parameters
    ----------
    filenames_array : list
        List of filenames to be removed from the specified directory.
    directory : str, optional
        Path to the directory containing the files to be removed.
        If not specified, the current directory is used.
    
    Returns
    -------
    None
    
    Notes
    -----
    This function attempts to remove each file in the provided list from the
    specified directory. If a file does not exist, the function silently
    continues without raising an error.
    
    This utility is necessary for cleaning DFDC files. 
    
    References
    ----------
    [1] Drela, M. and Youngren, H., AVL, http://web.mit.edu/drela/Public/web/avl
    """    	
    for f in filenames_array:
        try:
            os.remove(os.path.abspath(os.path.join(directory,f)))
        except OSError:
            pass 


# RCAIDE/Methods/Powertrain/Sources/Fuel_Cells/Larminie_Nodel/append_fuel_cell_conditions.py
# 
# 
# Created: Nov 2024, M. Clarke
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports 
import RCAIDE 
from RCAIDE.Framework.Mission.Common     import   Conditions

# pack imports 
from copy import  deepcopy

# ----------------------------------------------------------------------------------------------------------------------
# append_fuel_cell_conditions
# ----------------------------------------------------------------------------------------------------------------------  
def append_fuel_cell_conditions(fuel_cell_stack,segment,bus):
    """
    Appends the initial fuel_cell conditions. 

    Parameters
    ----------
    fuel_cell : fuel_cell
        The fuel_cell object containing cell properties and configuration.
    segment : MissionSegment
        The current mission segment. 
    bus : bus
        The electrical bus object.

    Returns
    ------- 

    Notes
    -----
    The function appends various fuel cell conditions in the `state` object, including: 
        - power                                      
        - voltage_under_load                         
        - voltage_open_circuit             
        - current_density                  
        - current                          
        - H2_mass_flow_rate
 
    References
    ---------- 
    """      
    
    ones_row = segment.state.ones_row  
                                             
    bus_conditions                                                                                 = segment.state.conditions.energy[bus.tag]        
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag]                                           = Conditions()
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].fuel_cell                                 = Conditions()
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].fuel_cell.inputs                          = Conditions()
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].fuel_cell.outputs                         = Conditions()

    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].power                                     = 0 * ones_row(1)
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].voltage_under_load                        = 0 * ones_row(1)
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].current                                   = 0 * ones_row(1)  
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].voltage_open_circuit                      = 0 * ones_row(1) 
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].fuel_cell.voltage_open_circuit            = 0 * ones_row(1)  
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].fuel_cell.voltage_under_load              = 0 * ones_row(1) 
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].fuel_cell.power                           = 0 * ones_row(1)  
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].fuel_cell.current                         = 0 * ones_row(1)  
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].fuel_cell.inlet_H2_mass_flow_rate         = 0 * ones_row(1)
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].fuel_cell.inlet_air_mass_flow_rate        = 0 * ones_row(1) 
    bus_conditions.fuel_cell_stacks[fuel_cell_stack.tag].H2_mass_flow_rate                         = 0 * ones_row(1) 
    
    # Conditions for recharging fuel_cell 
    if isinstance(segment,RCAIDE.Framework.Mission.Segments.Ground.Battery_Recharge):
        segment.state.conditions.energy.recharging  = True 
        segment.state.unknowns['recharge']          =  0* ones_row(1)  
        segment.state.residuals.network['recharge'] =  0* ones_row(1)
    elif type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Battery_Discharge:
        segment.state.conditions.energy.recharging   = False 
        segment.state.unknowns['discharge']          =  0* ones_row(1)  
        segment.state.residuals.network['discharge'] =  0* ones_row(1)     
    else:
        segment.state.conditions.energy.recharging  = False             
    return
 
def append_fuel_cell_segment_conditions(fuel_cell_stack, bus, conditions, segment): 
    """
    Sets the initial fuel cell energy at the start of each segment as the last point from the previous segment
    
    Parameters
    ----------
    fuel_cell_stack : fuel_cell_stack
        The fuel_cell_stack object containing cell properties and configuration.
    bus : bus
        The electrical bus object.
    conditions : MissionConditions
        The current conditions of the mission segment segment
    segment : MissionSegment
        The current mission segment. 

    Returns
    -------  
    """ 
    fuel_cell_conditions = conditions[bus.tag].fuel_cell_stacks[fuel_cell_stack.tag]
    if segment.state.initials:  
        fuel_cell_initials                                   = segment.state.initials.conditions.energy[bus.tag].fuel_cell_stacks[fuel_cell_stack.tag]    
        fuel_cell_conditions.temperature[:,0]                = fuel_cell_initials.temperature[-1,0]
        fuel_cell_conditions.cell.temperature[:,0]           = fuel_cell_initials.cell.temperature[-1,0]     
    return
  
def reuse_stored_fuel_cell_data(fuel_cell_stack,state,bus,stored_results_flag, stored_fuel_cell_stack_tag):
    '''Reuses results from one propulsor for identical fuel cells     
    ''' 
    state.conditions.energy[bus.tag].fuel_cell_stacks[fuel_cell_stack.tag] = deepcopy(state.conditions.energy[bus.tag].fuel_cell_stacks[stored_fuel_cell_stack_tag])
     
    return


# Carbon_Fiber_Honeycomb.py
#
# Created: Jul, 2017, J. Smart
# Modified: Apr, 2018, J. Smart

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from .Solid import Solid
from RCAIDE.Framework.Core import Units

#-------------------------------------------------------------------------------
# Carbon Fiber Honeycomb Core Solid Class
#-------------------------------------------------------------------------------

class Carbon_Fiber_Honeycomb(Solid):
    """
    A class representing carbon fiber honeycomb core material properties.

    Attributes
    ----------
    ultimate_tensile_strength : float
        Maximum tensile stress before failure in Pa (1e6)
    ultimate_shear_strength : float
        Maximum shear stress before failure in Pa (1e6)
    ultimate_bearing_strength : float
        Maximum bearing stress before failure in Pa (1e6)
    yield_tensile_strength : float
        Stress at which material begins to deform plastically in Pa (1e6)
    yield_shear_strength : float
        Shear stress at which material begins to deform plastically in Pa (1e6)
    yield_bearing_strength : float
        Bearing stress at which material begins to deform plastically in Pa (1e6)
    minimum_gage_thickness : float
        Minimum manufacturable thickness in m (6.5e-3)
    density : float
        Material density in kg/m³ (55.0)

    Notes
    -----
    This class implements material properties for carbon fiber honeycomb core materials
    typically used in sandwich composite structures. The low density and relatively high
    thickness reflect its primary use as a lightweight core material.

    **Definitions**
    
    'Ultimate Strength'
        The maximum stress that a material can withstand before failure
    
    'Yield Strength'
        The stress at which a material begins to deform plastically

    """

    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """

        self.ultimate_tensile_strength  = 1e6       * Units.Pa
        self.ultimate_shear_strength    = 1e6       * Units.Pa
        self.ultimate_bearing_strength  = 1e6       * Units.Pa
        self.yield_tensile_strength     = 1e6       * Units.Pa
        self.yield_shear_strength       = 1e6       * Units.Pa
        self.yield_bearing_strength     = 1e6       * Units.Pa
        self.minimum_gage_thickness     = 6.5e-3    * Units.m
        self.density                    = 55.       * Units['kg/(m**3)']



# RCAIDE/Methods/Energy/Propulsors/design_propeller.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
import RCAIDE
from RCAIDE.Framework.Core   import interp2d
from RCAIDE.Library.Methods.Geometry.Airfoil    import compute_airfoil_properties, compute_naca_4series, import_airfoil_geometry

# package imports 
import numpy as np
import scipy as sp 
from scipy.optimize import root 

# ----------------------------------------------------------------------------------------------------------------------  
#  Design Propeller
# ----------------------------------------------------------------------------------------------------------------------   
def design_propeller(prop, number_of_stations=20):
    """
    Optimizes propeller chord and twist distribution given input parameters.
    
    Parameters
    ----------
    prop : RCAIDE.Library.Components.Powertrain.Converters.Rotor
        Propeller component with the following attributes:
            - fidelity : str
                Analysis fidelity level
            - number_of_blades : int
                Number of blades on the propeller
            - tip_radius : float
                Tip radius of the propeller [m]
            - hub_radius : float
                Hub radius of the propeller [m]
            - cruise : Data
                Cruise conditions
                    - design_angular_velocity : float
                        Rotation rate [rad/s]
                    - design_freestream_velocity : float
                        Freestream velocity [m/s]
                    - design_Cl : float
                        Design lift coefficient
                    - design_altitude : float
                        Design altitude [m]
                    - design_thrust : float, optional
                        Design thrust [N] (specify either thrust or power)
                    - design_power : float, optional
                        Design power [W] (specify either thrust or power)
            - airfoils : dict
                Dictionary of airfoil objects
            - airfoil_polar_stations : list
                Indices of airfoil polars for each radial station
    number_of_stations : int, optional
        Number of radial stations for blade discretization, default 20
    
    Returns
    -------
    prop : RCAIDE.Library.Components.Powertrain.Converters.Rotor
        Propeller with optimized parameters:
            - cruise.design_power : float
                Design power [W]
            - cruise.design_thrust : float
                Design thrust [N]
            - cruise.design_torque : float
                Design torque [N·m]
            - max_thickness_distribution : array_like
                Maximum thickness distribution [m]
            - twist_distribution : array_like
                Blade twist distribution [rad]
            - chord_distribution : array_like
                Blade chord distribution [m]
            - radius_distribution : array_like
                Radial station positions [m]
            - cruise.design_power_coefficient : float
                Design power coefficient
            - cruise.design_thrust_coefficient : float
                Design thrust coefficient
            - mid_chord_alignment : array_like
                Mid-chord alignment [m]
            - thickness_to_chord : array_like
                Thickness-to-chord ratio
            - blade_solidity : float
                Blade solidity
    
    Notes
    -----
    This function implements the design methodology from "Design of Optimum Propellers"
    by Adkins and Liebeck to optimize propeller chord and twist distributions. It
    iteratively solves for the optimal circulation distribution that minimizes induced
    losses.
    
    The design process follows these steps:
        1. Calculate atmospheric properties at the design altitude
        2. Compute non-dimensional thrust or power coefficient
        3. Initialize the wake skew angle (zeta)
        4. Iteratively solve for the optimal circulation distribution:
            a. Compute the Prandtl momentum loss factor
            b. Determine the product of relative velocity and chord (Wc)
            c. Calculate Reynolds number and Mach number at each station
            d. Compute optimal angle of attack and drag coefficient
            e. Calculate the efficiency parameter (epsilon = Cd/Cl)
            f. Determine axial and tangential induction factors
            g. Compute chord and blade twist angle
            h. Calculate derivatives for thrust and power integrals
            i. Update the wake skew angle (zeta)
        5. Calculate final performance parameters (thrust, power, efficiency)
        6. Compute thickness distribution and blade solidity
    
    **Major Assumptions**
        * Either design thrust or design power must be specified (not both)
        * The design is optimized for a single operating condition
        * Airfoil performance is based on 2D characteristics
        * The wake is modeled with a constant skew angle
    
    **Theory**
    The method is based on the classical blade element momentum theory with
    modifications to account for wake rotation. The key parameter in the optimization
    is the wake skew angle (zeta), which relates to the induced velocities.
    
    The optimal circulation distribution minimizes induced losses while satisfying
    the thrust or power constraint. The method iteratively solves for this distribution
    by updating the wake skew angle until convergence.
    
    References
    ----------
    [1] Adkins, C.N. and Liebeck, R.H., "Design of Optimum Propellers", Journal of Propulsion and Power, Vol. 10, No. 5, 1994, pp. 676-682
    
    See Also
    --------
    RCAIDE.Library.Methods.Geometry.Airfoil.compute_airfoil_properties
    RCAIDE.Library.Methods.Geometry.Airfoil.compute_naca_4series
    RCAIDE.Library.Methods.Geometry.Airfoil.import_airfoil_geometry
    """
    if prop.fidelity == 'Blade_Element_Momentum_Theory_Helmholtz_Wake':
        # Unpack
        N            = number_of_stations       # this number determines the discretization of the propeller into stations
        B            = prop.number_of_blades
        R            = prop.tip_radius
        Rh           = prop.hub_radius
        omega        = prop.cruise.design_angular_velocity    # Rotation Rate in rad/s
        V            = prop.cruise.design_freestream_velocity # Freestream Velocity
        Cl           = prop.cruise.design_Cl                  # Design Lift Coefficient
        alt          = prop.cruise.design_altitude
        Thrust       = prop.cruise.design_thrust
        Power        = prop.cruise.design_power
        airfoils     = prop.airfoils 
        a_loc        = prop.airfoil_polar_stations

        if (Thrust == None) and (Power== None):
            raise AssertionError('Specify either design thrust or design power!')

        elif (Thrust!= None) and (Power!= None):
            raise AssertionError('Specify either design thrust or design power!') 

        if V == 0.0:
            V = 1E-6 

        # Calculate atmospheric properties
        atmosphere     = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data      = atmosphere.compute_values(alt) 
        T              = atmo_data.temperature[0]
        rho            = atmo_data.density[0]
        speed_of_sound = atmo_data.speed_of_sound[0]
        mu             = atmo_data.dynamic_viscosity[0]
        nu             = mu/rho

        # Nondimensional thrust
        if (Thrust!= None) and (Power == None):
            Tc = 2.*Thrust/(rho*(V*V)*np.pi*(R*R))     
            Pc = 0.0 

        elif (Thrust== None) and (Power != None):
            Tc = 0.0   
            Pc = 2.*Power/(rho*(V*V*V)*np.pi*(R*R))  

        tol   = 1e-10 # Convergence tolerance

        # Step 1, assume a zeta
        zeta = 0.1 # Assume to be small initially

        # Step 2, determine F and phi at each blade station 
        chi0    = Rh/R # Where the propeller blade actually starts
        chi     = np.linspace(chi0,1,N+1) # Vector of nondimensional radii
        chi     = chi[0:N]
        lamda   = V/(omega*R)             # Speed ratio
        r       = chi*R                   # Radial coordinate
        x       = omega*r/V               # Nondimensional distance
        diff    = 1.0                     # Difference between zetas
        n       = omega/(2*np.pi)         # Cycles per second
        D       = 2.*R  
        c       = 0.2 * np.ones_like(chi)

        # if user defines airfoil, check dimension of stations
        num_airfoils = len(airfoils.keys())
        if num_airfoils>0:
            if len(a_loc) != N:
                raise AssertionError('\nDimension of airfoil sections must be equal to number of stations on propeller') 

            for _,airfoil in enumerate(airfoils):  
                if airfoil.geometry == None: # first, if airfoil geometry data not defined, import from geoemtry files
                    if type(airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil: # check if naca 4 series of airfoil from datafile
                        airfoil.geometry = compute_naca_4series(airfoil.NACA_4_Series_code,airfoil.number_of_points)
                    else:
                        airfoil.geometry = import_airfoil_geometry(airfoil.coordinate_file,airfoil.number_of_points) 

                if airfoil.polars == None: # compute airfoil polars for airfoils
                    airfoil.polars = compute_airfoil_properties(airfoil.geometry, airfoil_polar_files= airfoil.polar_files) 
        else:
            print('\nDefaulting to scaled DAE51') 

        while diff>tol:      
            # assign chord distribution
            prop.chord_distribution = c  

            #Things that need a loop
            Tcnew   = Tc     
            tanphit = lamda*(1.+zeta/2.)   # Tangent of the flow angle at the tip
            phit    = np.arctan(tanphit)   # Flow angle at the tip
            tanphi  = tanphit/chi          # Flow angle at every station
            f       = (B/2.)*(1.-chi)/np.sin(phit) 
            F       = (2./np.pi)*np.arccos(np.exp(-f)) #Prandtl momentum loss factor
            phi     = np.arctan(tanphi)    # Flow angle at every station

            #Step 3, determine the product Wc, and RE
            G       = F*x*np.cos(phi)*np.sin(phi) #Circulation function
            Wc      = 4.*np.pi*lamda*G*V*R*zeta/(Cl*B)
            Ma      = Wc/speed_of_sound
            RE      = Wc/nu

            if num_airfoils>0:
                # assign initial values 
                alpha0   = np.ones(N)*0.05

                # solve for optimal alpha to meet design Cl target
                sol      = root(objective, x0 = alpha0 , args=(airfoils,a_loc,RE,Cl,N))
                alpha    = sol.x

                # query surrogate for sectional Cls at stations 
                Cdval    = np.zeros_like(RE) 
                for j,airfoil in enumerate(airfoils):                   
                    pd          = airfoil.polars
                    Cdval_af    = interp2d(RE,alpha,pd.reynolds_numbers, pd.angle_of_attacks, pd.drag_coefficients)
                    locs        = np.where(np.array(a_loc) == j )
                    Cdval[locs] = Cdval_af[locs]    

            else:    
                Cdval   = (0.108*(Cl**4)-0.2612*(Cl**3)+0.181*(Cl**2)-0.0139*Cl+0.0278)*((50000./RE)**0.2)
                alpha   = Cl/(2.*np.pi)

            #More Cd scaling from Mach from AA241ab notes for turbulent skin friction
            Tw_Tinf = 1. + 1.78*(Ma**2)
            Tp_Tinf = 1. + 0.035*(Ma**2) + 0.45*(Tw_Tinf-1.)
            Tp      = Tp_Tinf*T
            Rp_Rinf = (Tp_Tinf**2.5)*(Tp+110.4)/(T+110.4) 
            Cd      = ((1/Tp_Tinf)*(1/Rp_Rinf)**0.2)*Cdval

            #Step 5, change Cl and repeat steps 3 and 4 until epsilon is minimized 
            epsilon = Cd/Cl  

            #Step 6, determine a and a', and W 
            a       = (zeta/2.)*(np.cos(phi)**2.)*(1.-epsilon*np.tan(phi)) 
            W       = V*(1.+a)/np.sin(phi)

            #Step 7, compute the chord length and blade twist angle  
            c       = Wc/W
            beta    = alpha + phi # Blade twist angle

            #Step 8, determine 4 derivatives in I and J 
            Iprime1 = 4.*chi*G*(1.-epsilon*np.tan(phi))
            Iprime2 = lamda*(Iprime1/(2.*chi))*(1.+epsilon/np.tan(phi)
                                                )*np.sin(phi)*np.cos(phi)
            Jprime1 = 4.*chi*G*(1.+epsilon/np.tan(phi))
            Jprime2 = (Jprime1/2.)*(1.-epsilon*np.tan(phi))*(np.cos(phi)**2.) 
            dchi    = (chi[1]-chi[0])*np.ones_like(Jprime1)

            #Integrate derivatives from chi=chi0 to chi=1 
            I1      = np.dot(Iprime1,dchi)
            I2      = np.dot(Iprime2,dchi)
            J1      = np.dot(Jprime1,dchi)
            J2      = np.dot(Jprime2,dchi)        

            #Step 9, determine zeta and and Pc or zeta and Tc 
            if (Pc==0.)&(Tc!=0.): 
                #First Case, Thrust is given
                #Check to see if Tc is feasible, otherwise try a reasonable number
                if Tcnew>=I2*(I1/(2.*I2))**2.:
                    Tcnew = I2*(I1/(2.*I2))**2.
                zetan    = (I1/(2.*I2)) - ((I1/(2.*I2))**2.-Tcnew/I2)**0.5

            elif (Pc!=0.)&(Tc==0.): 
                #Second Case, Thrust is given
                zetan    = -(J1/(J2*2.)) + ((J1/(J2*2.))**2.+Pc/J2)**0.5 

            #Step 10, repeat starting at step 2 with the new zeta
            diff = abs(zeta-zetan)

            zeta = zetan

        # Step 11, determine propeller efficiency etc...
        if (Pc==0.)&(Tc!=0.): 
            if Tcnew>=I2*(I1/(2.*I2))**2.:
                Tcnew = I2*(I1/(2.*I2))**2.
                print('Tc infeasible, reset to:')
                print(Tcnew)        
            #First Case, Thrust is given
            zeta    = (I1/(2.*I2)) - ((I1/(2.*I2))**2.-Tcnew/I2)**0.5
            Pc      = J1*zeta + J2*(zeta**2.)
            Tc      = I1*zeta - I2*(zeta**2.)

        elif (Pc!=0.)&(Tc==0.): 
            #Second Case, Thrust is given
            zeta    = -(J1/(2.*J2)) + ((J1/(2.*J2))**2.+Pc/J2)**0.5
            Tc      = I1*zeta - I2*(zeta**2.)
            Pc      = J1*zeta + J2*(zeta**2.) 

        # Calculate mid-chord alignment angle, MCA
        # This is the distance from the mid chord to the line axis out of the center of the blade
        # In this case the 1/4 chords are all aligned 
        MCA    = c/4. - c[0]/4.

        Thrust = Tc*rho*(V**2)*np.pi*(R**2)/2
        Power  = Pc*rho*(V**3)*np.pi*(R**2)/2 
        Ct     = Thrust/(rho*(n*n)*(D*D*D*D))
        Cp     = Power/(rho*(n*n*n)*(D*D*D*D*D))  

        # compute max thickness distribution  
        t_max  = np.zeros(N)    
        t_c    = np.zeros(N)   
        if num_airfoils>0:
            for j,airfoil in enumerate(airfoils): 
                a_geo         = airfoil.geometry
                locs          = np.where(np.array(a_loc) == j )
                t_max[locs]   = a_geo.max_thickness*c[locs] 
                t_c[locs]     = a_geo.thickness_to_chord 
        else:     
            c_blade = np.repeat(np.atleast_2d(np.linspace(0,1,N)),N, axis = 0)* np.repeat(np.atleast_2d(c).T,N, axis = 1)
            t       = (5*c_blade)*(0.2969*np.sqrt(c_blade) - 0.1260*c_blade - 0.3516*(c_blade**2) + 0.2843*(c_blade**3) - 0.1015*(c_blade**4)) # local thickness distribution
            t_max   = np.max(t,axis = 1) 
            t_c     = np.max(t,axis = 1) /c  

        # Nondimensional thrust
        if prop.cruise.design_power == None: 
            prop.cruise.design_power = Power[0]        
        elif prop.cruise.design_thrust == None: 
            prop.cruise.design_thrust = Thrust[0]       

        # blade solidity
        r          = chi*R                     
        blade_area = sp.integrate.cumulative_trapezoid(B*c, r-r[0])
        sigma      = blade_area[-1]/(np.pi*R**2)   

        prop.cruise.design_torque                   = Power[0]/omega
        prop.max_thickness_distribution             = t_max
        prop.twist_distribution                     = beta
        prop.chord_distribution                     = c
        prop.radius_distribution                    = r
        prop.number_of_blades                       = int(B)
        prop.cruise.design_power_coefficient        = Cp
        prop.cruise.design_thrust_coefficient       = Ct
        prop.mid_chord_alignment                    = MCA
        prop.thickness_to_chord                     = t_c
        prop.blade_solidity                         = sigma  

    return prop

    
def objective(x,airfoils,a_loc,RE,Cl,N):
    # query surrogate for sectional Cls at stations 
    Cl_vals          = np.zeros(N)       
    for j,airfoil in enumerate(airfoils): 
        pd            = airfoil.polars
        Cl_af         = interp2d(RE,x,pd.reynolds_numbers, pd.angle_of_attacks, pd.lift_coefficients)
        locs          = np.where(np.array(a_loc) == j )
        Cl_vals[locs] = Cl_af[locs] 
        
    # compute Cl residual    
    Cl_residuals = Cl_vals - Cl 
    return  Cl_residuals 



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Raymer/compute_landing_gear_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE
from RCAIDE.Framework.Core    import Units ,  Data 
 
# ----------------------------------------------------------------------------------------------------------------------
#  Horizontal Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_landing_gear_weight(vehicle):
    """
    Calculates the weight of main and nose landing gear for transport aircraft using Raymer's method.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure containing:
            - design_range : float
                Design range of aircraft [m]
            - mass_properties.max_takeoff : float
                Maximum takeoff weight [kg]
            - systems.accessories : str
                Aircraft type ('short-range', 'commuter', 'medium-range',
                'long-range', 'sst', 'cargo')
            - landing_gears : list
                List of landing gear components containing:
                    - strut_length : float
                        Length of strut [m]
                    - wheels : int
                        Number of wheels per strut

    Returns
    -------
    output : Data()
        Landing gear weights:
            - main : float
                Weight of main landing gear [kg]
            - nose : float
                Weight of nose landing gear [kg]

    Notes
    -----
    This method implements Raymer's empirical correlations for transport aircraft
    landing gear weight estimation. Separate correlations are used for main and
    nose gear.

    **Major Assumptions**
        * Gear load factor = 3
        * Number of main gear shock struts = 2
        * Stall speed assumes max Cl of 2.5, density of 1.225 kg/m^3
        * All main landing gear is assumed to be of the same type
        * All nose landing gear is assumed to be of the same type

    **Theory**
    The landing gear weights are calculated using:
    .. math::
        W_{main} = 0.0106K_{mp}W_{l}^{0.888}N_l^{0.25}L_m^{0.4}N_{mw}^{0.321}N_{mss}^{-0.5}V_{stall}^{0.1}

    .. math::
        W_{nose} = 0.032K_{np}W_{l}^{0.646}N_l^{0.2}L_n^{0.5}N_{nw}^{0.45}

    where:
        * :math:`W_l` is landing weight
        * :math:`N_l` is ultimate landing load factor
        * :math:`L_m, L_n` are strut lengths
        * :math:`N_{mw}, N_{nw}` are number of wheels
        * :math:`N_{mss}` is number of main gear shock struts

    References
    ----------
    [1] Raymer, D., "Aircraft Design: A Conceptual Approach", AIAA 
        Education Series, 2018. 

    See Also
    --------
    RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Transport.Raymer.compute_operating_empty_weight
    """
    Kmp         = 1  # assuming not a kneeling gear
    if vehicle.systems.accessories == "sst":
        RFACT   = 0.00009
    else:
        RFACT   = 0.00004
    DESRNG      = vehicle.flight_envelope.design_range / Units.nmi  # Design range in nautical miles
    WLDG        = vehicle.mass_properties.max_takeoff / Units.lbs * (1 - RFACT * DESRNG)
    Ngear       = 3  # gear load factor, usually around 3
    Nl          = Ngear * 1.5  # ultimate landing load factor
      
    Nmw  = 0
    Lm   = 0 
    Ln   = 0
    Nnw  = 0
    Nmss = 2  # number of main gear shock struts assumed to be 2
    for landing_gear in  vehicle.landing_gears:
        if isinstance(landing_gear,RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear):
            Nmw  = landing_gear.wheels * Nmss    
            Lm   = landing_gear.strut_length / Units.inch 
        elif isinstance(landing_gear,RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear): 
            Ln          = landing_gear.strut_length / Units.inch
            Nnw         = landing_gear.wheels  
        
    Vstall      = (2 * vehicle.mass_properties.max_takeoff * 9.81 / (vehicle.wings.main_wing.areas.reference * 1.225 * 2.5)) ** 0.5 # Assumes max Cl of 2.5, density of 1.225 kg/m^3
    Knp         = 1  # assuming not a kneeling gear
    W_main_landing_gear = 0.0106 * Kmp * WLDG ** 0.888 * Nl ** 0.25 * Lm ** 0.4 * Nmw ** 0.321 * Nmss ** (-0.5) * Vstall ** 0.1
    W_nose_landing_gear = 0.032 * Knp * WLDG ** 0.646 * Nl ** 0.2 * Ln ** 0.5 * Nnw ** 0.45

    output          = Data()
    output.main     = W_main_landing_gear * Units.lbs
    output.nose     = W_nose_landing_gear * Units.lbs
    return output